/* 
* Copyright (C) 2015, Adrian Jarabo (http://giga.cps.unizar.es/~ajarabo/)
*
* Permission is hereby granted, free of charge, to any person obtaining
* a copy of this software and associated documentation files (the "Software"),
* to deal in the Software without restriction, including without limitation
* the rights to use, copy, modify, merge, publish, distribute, sublicense,
* and/or sell copies of the Software, and to permit persons to whom
* the Software is furnished to do so, subject to the following conditions:

* The above copyright notice and this permission notice shall be included
* in all copies or substantial portions of the Software.

* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
* EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
* MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
* IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
* DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
* TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
* OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#ifndef _SPECTRUM_H_
#define _SPECTRUM_H_

#include "Utils/globals.h"
#include "Color/Color.h"
#include "Utils/Utils.h"
#include <math.h>

template<int N, int WL_MIN=380, int WL_MAX=780>
class tSpectrum
{
public:
	tSpectrum(const Real p=0.)
	{
		for( int i=0; i<N; ++i)
			data[i] = p;
	}
	tSpectrum(const tSpectrum<N,WL_MIN,WL_MAX> &s)
	{
		for( int i=0; i<N; ++i)
			data[i] = s.data[i];
	}
	/**	This function transforms general RGB into spectral
		color. It bases on:
		Smits, B. "An RGB-to-spectrum conversion for 
		reflectances". In Journal of Graphics Tools Vol.4(4).
		http://www.cs.utah.edu/~bes/papers/color/		*/	
	//	[NEED TO BE DONE]								
	tSpectrum(const Real r, const Real g, const Real b)
	{
	}
	tSpectrum( const std::vector<Real> &lambda, const Real *values )
	{
		extract_values(lambda, values);
	}

	
	// Access operations
	inline Real operator[](const int dimension) const
	{
		#ifdef _SAFE_CHECK_		
			if (dimension > N-1 || dimension < 0) throw("Out-of-range");
		#endif	
		return data[dimension];	
	}
	inline Real &operator[](const int dimension)
	{
		#ifdef _SAFE_CHECK_		
			if (dimension > N-1 || dimension < 0) throw("Out-of-range");
		#endif	
		return data[dimension];	
	} 
	inline Real get_max()const
	{
		Real value = -1.;
		for(int i=0; i<N; ++i )
		{
			if( value < data[i] )
				value = data[i];
		}
		return value;
	}

	inline Real get_min()const
	{
		Real value = std::numeric_limits<Real>::infinity();
		for(int i=0; i<N; ++i )
		{
			if( value > data[i] )
				value = data[i];
		}
		return value;
	}

	
	inline int get_spectral_samples()const{ return N; }
	virtual void get_lambda( Real *lambda )const
	{
		
		Real delta_lambda = static_cast<Real>(WL_MAX-WL_MIN);
		
		if( N == 1 ) delta_lambda = 0; 
		else delta_lambda /= static_cast<Real>(N-1);

		Real current_lambda = WL_MIN;
		for( int i=0; i<N; ++i, current_lambda += delta_lambda )
			lambda[i] = current_lambda;

	}

	virtual void extract_values( const std::vector<Real> &lambda, const Real *values) 
	{
		Real delta_lambda = static_cast<Real>(WL_MAX-WL_MIN);
		
		if( N == 1 ) delta_lambda = 0; 
		else delta_lambda /= static_cast<Real>(N-1);

		Real current_lambda = WL_MIN;
		for( int i=0; i<N; ++i, current_lambda += delta_lambda )
			data[i] = Utils::interpolate(lambda, values, current_lambda);

	}
	
	// Comparison 
	inline bool operator==(const tSpectrum<N,WL_MIN,WL_MAX> &v)const
	{
		for( int i=0; i<N; ++i)
			if( data[i] != v.data[i] )
				return false;

		return true;
	}

	inline tSpectrum<N,WL_MIN,WL_MAX> max(const tSpectrum<N,WL_MIN,WL_MAX> &v)const
	{
		tSpectrum<N,WL_MIN,WL_MAX> nspectrum;
		for( int i=0; i<N; ++i)
			nspectrum[i] = std::max(data[i], v.data[i]);

		return nspectrum;
	}

	
	inline tSpectrum<N,WL_MIN,WL_MAX> min(const tSpectrum<N,WL_MIN,WL_MAX> &v)const
	{
		tSpectrum<N,WL_MIN,WL_MAX> nspectrum;
		for( int i=0; i<N; ++i)
			nspectrum[i] = std::min(data[i], v.data[i]);

		return nspectrum;
	}


	
	// Operations
	inline Real sum()const
	{
		Real value = 0.;
		for(int i=0; i<N; ++i )
			value += data[i];

		return value;
	}

	inline Real avg()const
	{
		Real value = 0.;
		for(int i=0; i<N; ++i )
			value += data[i];

		return value/static_cast<Real>(N);
	}
	inline bool isZero(void) const
	{
		for( int i=0; i<N; ++i )
			if( data[i] > 0. )
				return false;

		return true;
	}
	inline Real length() const
	{
		return sqrtf(length2());
	}
	inline Real length2() const
	{
		Real value = 0.;
		for(int i=0; i<N; ++i )
			value += data[i]*data[i];
		
		return value;
	}
	tSpectrum<N,WL_MIN,WL_MAX> exp()const
	{
		tSpectrum<N,WL_MIN,WL_MAX> nspectrum;
		for( int i=0; i<N; ++i)
			nspectrum[i] = expf(data[i]);

		return nspectrum;
	}

	tSpectrum<N,WL_MIN,WL_MAX> operator-()const
	{
		tSpectrum<N,WL_MIN,WL_MAX> nspectrum;
		for( int i=0; i<N; ++i)
			nspectrum[i] = -data[i];

		return nspectrum;
	}
	tSpectrum<N,WL_MIN,WL_MAX> operator+(const tSpectrum<N,WL_MIN,WL_MAX>& v) const
	{
		tSpectrum<N,WL_MIN,WL_MAX> nspectrum;
		for( int i=0; i<N; ++i)
			nspectrum[i] = data[i]+v[i];

		return nspectrum;
	}
	tSpectrum<N,WL_MIN,WL_MAX> operator-(const tSpectrum<N,WL_MIN,WL_MAX>& v) const
	{
		tSpectrum<N,WL_MIN,WL_MAX> nspectrum;
		for( int i=0; i<N; ++i)
			nspectrum[i] = data[i]-v[i];

		return nspectrum;
	}
	tSpectrum<N,WL_MIN,WL_MAX> operator*(const Real f) const
	{
		tSpectrum<N,WL_MIN,WL_MAX> nspectrum;
		for( int i=0; i<N; ++i)
			nspectrum[i] = data[i]*f;

		return nspectrum;
	}
	tSpectrum<N,WL_MIN,WL_MAX> operator*(const tSpectrum<N,WL_MIN,WL_MAX>& v) const
	{
		tSpectrum<N,WL_MIN,WL_MAX> nspectrum;
		for( int i=0; i<N; ++i)
			nspectrum[i] = data[i]*v[i];

		return nspectrum;
	}
	tSpectrum<N,WL_MIN,WL_MAX> operator/(const Real f) const
	{
		tSpectrum<N,WL_MIN,WL_MAX> nspectrum;
		Real inv_f = 1./f;
		for( int i=0; i<N; ++i)
			nspectrum[i] = data[i]*inv_f;

		return nspectrum;
	}
	tSpectrum<N,WL_MIN,WL_MAX> operator/(const tSpectrum<N,WL_MIN,WL_MAX>& v) const
	{
		tSpectrum<N,WL_MIN,WL_MAX> nspectrum;
		for( int i=0; i<N; ++i)
			nspectrum[i] = data[i]/v[i];

		return nspectrum;
	}

	const tSpectrum<N,WL_MIN,WL_MAX> &operator+=(const tSpectrum<N,WL_MIN,WL_MAX>& v)
	{
		for( int i=0; i<N; ++i)
			data[i]+=v[i];

		return *this;
	}
	const tSpectrum<N,WL_MIN,WL_MAX> &operator-=(const tSpectrum<N,WL_MIN,WL_MAX>& v)
	{
		for( int i=0; i<N; ++i)
			data[i]-=v[i];

		return *this;
	}
	const tSpectrum<N,WL_MIN,WL_MAX> &operator*=(const tSpectrum<N,WL_MIN,WL_MAX>& v)
	{
		for( int i=0; i<N; ++i)
			data[i]*=v[i];

		return *this;
	}
	const tSpectrum<N,WL_MIN,WL_MAX> &operator/=(const tSpectrum<N,WL_MIN,WL_MAX>& v)
	{
		for( int i=0; i<N; ++i)
			data[i]/=v[i];

		return *this;
	}

	const tSpectrum<N,WL_MIN,WL_MAX> &operator*=(const Real f)
	{
		for( int i=0; i<N; ++i)
			data[i]*=f;

		return *this;
	}
	const tSpectrum<N,WL_MIN,WL_MAX> &operator/=(const Real f)
	{
		Real inv_f = 1./f;
		for( int i=0; i<N; ++i)
			data[i]*=inv_f;

		return *this;
	}
	
	// Assignation operators
	const tSpectrum<N,WL_MIN,WL_MAX> &operator=(const tSpectrum<N,WL_MIN,WL_MAX>& v)
	{
		for( int i=0; i<N; ++i)
			data[i]=v[i];

		return *this;
	}

	const tSpectrum<N,WL_MIN,WL_MAX> &operator=(const Real f)
	{
		for( int i=0; i<N; ++i)
			data[i]=f;

		return *this;
	}
	// Transformation to RGB functions
	Vector3 to_rgb()const;

	// Class member
	Real data[N];

}; //tSpectrum

template<int N, int WL_MIN, int WL_MAX>
inline tSpectrum<N,WL_MIN,WL_MAX> operator+(const Real f, const tSpectrum<N,WL_MIN,WL_MAX> &v)
{
	return v+f;
}
template<int N, int WL_MIN, int WL_MAX>
inline tSpectrum<N,WL_MIN,WL_MAX> operator-(const Real f, const tSpectrum<N,WL_MIN,WL_MAX> &v)
{
	return tSpectrum<N,WL_MIN,WL_MAX>(f)-v;
}
template<int N, int WL_MIN, int WL_MAX>
inline tSpectrum<N,WL_MIN,WL_MAX> operator*(const Real f, const tSpectrum<N,WL_MIN,WL_MAX> &v)
{
	return v*f;
}
template<int N, int WL_MIN, int WL_MAX>
inline tSpectrum<N,WL_MIN,WL_MAX> operator/(const Real f, const tSpectrum<N,WL_MIN,WL_MAX> &v)
{
	return tSpectrum<N,WL_MIN,WL_MAX>(f)/v;
}
template<int N, int WL_MIN, int WL_MAX>
inline tSpectrum<N,WL_MIN,WL_MAX> exp(const tSpectrum<N,WL_MIN,WL_MAX> &v)
{
	return v.exp();
}
template<int N, int WL_MIN, int WL_MAX>
inline tSpectrum<N,WL_MIN,WL_MAX> max(const tSpectrum<N,WL_MIN,WL_MAX> &v1, const tSpectrum<N,WL_MIN,WL_MAX> &v2)
{
	return v1.max(v2)
}
template<int N, int WL_MIN, int WL_MAX>
inline tSpectrum<N,WL_MIN,WL_MAX> min(const tSpectrum<N,WL_MIN,WL_MAX> &v1, const tSpectrum<N,WL_MIN,WL_MAX> &v2)
{
	return v1.min(v2)
}


//------------------------------------------------------------------------
// To RGB
template<>
Vector3 tSpectrum<81,380,780>::to_rgb()const
{
	Vector3 xyz;
	for( int i=0; i<81; ++i)
		xyz += Vector3(Color::WL2XYZ[i][0]*data[i],Color::WL2XYZ[i][1]*data[i],Color::WL2XYZ[i][2]*data[i]); 

	return Color::xyz2rgb(xyz);
}


// For now, need to be done... Oh well...
template<int N,int WL_MIN,int WL_MAX>
Vector3 tSpectrum<N,WL_MIN,WL_MAX>::to_rgb()const
{
	//Vector3 xyz;

	//int step_size = N / (WL_MAX-WL_MIN);
	//int init = WL_MIN + 
	//Real delta = static_cast<Real>(DELTA_SAMPLES_SPECTRUM)/static_cast<Real>(step_size);

	//
	//Real delta_wl = static_cast<Real>(N)/static_cast<Real>(WL_MAX-WL_MIN);
	//Real current = static_cast<Real>(WL_MIN);
	//

	//for( int i=0; i<N; ++i, current += delta_wl )
	//	
	//return Vector3(data[0],data[1],data[2]);
}



class Spectrum3: public tSpectrum<3,380,780>
{ 
public:
	Spectrum3(const Real p=0.):tSpectrum<3,380,780>(0)
	{
		data[0] =  data[1] = data[2] = p;
	}
	Spectrum3(const tSpectrum<3,380,780> &s)
	{
		for( int i=0; i<3; ++i)
			data[i] = s.data[i];
	}
	// Need to be done, base on: http://www.cs.utah.edu/~bes/papers/color/
	Spectrum3(const Real r, const Real g, const Real b)
	{
		data[0] = r; data[1] = g; data[2] = b;
	}
	
	Spectrum3( const std::vector<Real> &lambda, const Real *values )
	{
		extract_values(lambda, values);
	}

	Vector3 to_rgb()const
	{
		return Vector3(data[0],data[1],data[2]);
	}

	virtual void get_lambda( Real *lambda )const
	{
		lambda[0] = 650;
		lambda[1] =	520;
		lambda[2] =	465;
	}
	virtual void extract_values( const std::vector<Real> &lambda, const Real *values) 
	{
		data[0] = Utils::interpolate(lambda, values, 650.f);
		data[1] = Utils::interpolate(lambda, values, 520.f);
		data[2] = Utils::interpolate(lambda, values, 465.f);	
	}
}; //Spectrum3

class Spectrum1: public tSpectrum<1,380,380>
{ 
public:
	Spectrum1(const Real p=0.):tSpectrum<1,380,380>(0)
	{
		data[0] = p;
	}
	Spectrum1(const tSpectrum<1,380,380> &s)
	{
		data[0] = s.data[0];
	}
	/** Transforms RGB to luminances using the values given
		by Stokes et al. 1996. "A Standard Default Color Space 
		for the Internet - sRGB".
		http://www.w3.org/Graphics/Color/sRGB			*/
	Spectrum1(const Real r, const Real g, const Real b)
	{
		data[0] = 0.2126*r + 0.7152*g + 0.0722*b;
	}
	
	Vector3 to_rgb()const
	{
		return Vector3(data[0]);
	}
}; //Spectrum3

typedef Spectrum3 Spectrum;

#endif //_SPECTRUM_H_