/* 
* Copyright (C) 2015, Adrian Jarabo (http://giga.cps.unizar.es/~ajarabo/)
*
* Permission is hereby granted, free of charge, to any person obtaining
* a copy of this software and associated documentation files (the "Software"),
* to deal in the Software without restriction, including without limitation
* the rights to use, copy, modify, merge, publish, distribute, sublicense,
* and/or sell copies of the Software, and to permit persons to whom
* the Software is furnished to do so, subject to the following conditions:

* The above copyright notice and this permission notice shall be included
* in all copies or substantial portions of the Software.

* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
* EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
* MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
* IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
* DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
* TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
* OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#ifndef _STREAK_CAMERA_FILM_H_
#define _STREAK_CAMERA_FILM_H_

#include "Film/Film.h"
#include "Filter/UberFilter.h"
#include <vector>

class StreakCameraFilm: public Film
{
	int m_nb_available_slices, m_first_available_slice, m_last_available_slice;
	int m_time_resolution;
	Real m_exposure_time;
	Real m_offset;
	
	bool camera_unwarp;
	Components components;

	Filter *time_filter;
	std::vector<Real*> m_available_slices,m_available_slices_variance,m_available_slices_nsamples;
	std::string m_name, m_extension;


	void fix_and_store_slice(const int y)
	{
		Imaging::Image out_image(m_time_resolution, width);
		Imaging::Image out_image_var(m_time_resolution, width);
		Imaging::Image out_image_ns(m_time_resolution, width);

		// First, fix the slice (which basicaly means that we are normalizing the 
		// values on it!
		Real *curr_slice = m_available_slices[y-m_first_available_slice];
		Real *curr_slice_var = (components & VARIANCE)?m_available_slices_variance[y-m_first_available_slice]:NULL;	
		Real *curr_slice_ns = (components & VARIANCE)?m_available_slices_nsamples[y-m_first_available_slice]:NULL;	

		//int index_w = (y-m_first_available_slice)*width;
		int index_w = (y)*width;
		int index = 0;
		for(int i=0; i<width; ++i)
		{
			if( weight[index_w] )
				for(int t = 0; t < m_time_resolution; ++t)
				{
					Real inv_weight = 1./weight[index_w];
					
					out_image.set(t,i, Imaging::RGBColor(curr_slice[index  ]*inv_weight,
														 curr_slice[index+1]*inv_weight,
														 curr_slice[index+2]*inv_weight));
					if(components & VARIANCE)
					{
						out_image_var.set(t,i, Imaging::RGBColor(curr_slice_var[index  ]*inv_weight,
														 curr_slice_var[index+1]*inv_weight,
														 curr_slice_var[index+2]*inv_weight));
						out_image_ns.set(t,i, Imaging::RGBColor(curr_slice_ns[index  ]*inv_weight,
														 curr_slice_ns[index+1]*inv_weight,
														 curr_slice_ns[index+2]*inv_weight));
					}
					index+=3;
				}
			else
				index+=m_time_resolution*3;

			++index_w; 
		}

		//And after that, store it!
		char name[1024];
		char name_var[1024];
		char name_ns[1024];
		
		sprintf(name, "%s_%04d.%s", m_name.c_str(), y, m_extension.c_str());
        
        //workaround for hdr images less than 8 pixels wide (matlab fails reading them) => store them transposed 
        if (out_image.width() < 8) out_image.transpose();
        
		out_image.save(name);
		
		if (curr_slice_var)
		{ 
			sprintf(name_var, "%s_%04d_variance.%s", m_name.c_str(), y, m_extension.c_str());
            if (out_image_var.width() < 8) out_image_var.transpose();
			out_image_var.save(name_var);
		}

		if (curr_slice_ns)
		{ 
			sprintf(name_ns, "%s_%04d_nsamples.%s", m_name.c_str(), y, m_extension.c_str());
            if (out_image_ns.width() < 8) out_image_ns.transpose();
			out_image_ns.save(name_ns);
		}
		
	}
	void get_film_samples(const Real x, const unsigned int d, std::list<Real> &samples)const
	{
		if( d < 2 )
			Film::get_film_samples(x, d, samples );
		else
		{
			samples.clear();

			int size_kernel = filter->get_size();
			int half_size = size_kernel / 2;
			int ix = static_cast<int>(x);
			
			int init_x = x-half_size, end_x = x+half_size;

			init_x = (init_x < 0 )?0:init_x;
			end_x = (end_x < m_time_resolution)?end_x:(m_time_resolution-1);

			for( int i=init_x; i<=end_x; ++i )
				samples.push_back(i+.5);
		}
	}

public:
	StreakCameraFilm(int w, int h, int t, Real exposure, bool _camera_unwarp = false, Components comp = RADIANCE)
		:Film(w,h,comp), m_time_resolution(t), m_exposure_time(exposure), m_nb_available_slices((filter)?filter->get_size():1),
		m_first_available_slice(0), m_last_available_slice(m_nb_available_slices-1), m_name("frame"), m_extension("hdr"), camera_unwarp(false),
		time_filter(filter->get_subfilter(2)), components(comp), m_offset(0)
	{	
		if( t && w )
		{
			for(int i=0; i<m_nb_available_slices; ++i)
				m_available_slices.push_back(new Real[t*w*3]());

			if(components & VARIANCE)
			{
				for(int i=0; i<m_nb_available_slices; ++i)
				{
					m_available_slices_variance.push_back(new Real[t*w*3]());
					m_available_slices_nsamples.push_back(new Real[t*w*3]());
				}
			}

		}
	}

	StreakCameraFilm(int w, int h, int t, Real exposure, Filter *_filter, Components comp = RADIANCE )
		:Film(w,h,_filter,comp), m_time_resolution(t), m_exposure_time(exposure), m_nb_available_slices((filter)?filter->get_size():1),
		m_first_available_slice(0), m_last_available_slice(m_nb_available_slices-1), m_name("frame"), m_extension("hdr"),
		camera_unwarp(false), time_filter(filter->get_subfilter(2)), components(comp), m_offset(0)
	{	
		if( t && w )
		{
			for(int i=0; i<m_nb_available_slices; ++i)
				m_available_slices.push_back(new Real[t*w*3]());
			if(components & VARIANCE)
			{
				for(int i=0; i<m_nb_available_slices; ++i)
				{
					m_available_slices_variance.push_back(new Real[t*w*3]());
					m_available_slices_nsamples.push_back(new Real[t*w*3]());
				}
			}
		}
	}

	StreakCameraFilm(int w, int h, int t, Real exposure, Filter *_filter, bool _camera_unwarp = false, Components comp = RADIANCE )
		:Film(w,h,_filter,comp), m_time_resolution(t), m_exposure_time(exposure), m_nb_available_slices((filter)?filter->get_size():1),
		m_first_available_slice(0), m_last_available_slice(m_nb_available_slices-1), m_name("frame"), m_extension("hdr"),
		camera_unwarp(_camera_unwarp), time_filter(filter->get_subfilter(2)), components(comp), m_offset(0)
	{	
		if( t && w )
		{
			for(int i=0; i<m_nb_available_slices; ++i)
				m_available_slices.push_back(new Real[t*w*3]());
			//for (int k = 0; k < t*w*3; k++)
			//		printf("%f\t", m_available_slices[0][k]);
			if(components & VARIANCE)
			{
				for(int i=0; i<m_nb_available_slices; ++i)
				{
					m_available_slices_variance.push_back(new Real[t*w*3]());
					m_available_slices_nsamples.push_back(new Real[t*w*3]());
				}
			}
		}
	}
	
	~StreakCameraFilm()
	{
		clear();
	}

	// Returns time resolution of StreakFilm
	virtual int get_time_resolution() const {return m_time_resolution;}
	//Allows setting the name of the streak images
	virtual void set_name(const char *name, const char *ext) {m_name = std::string(name); m_extension = std::string(ext);}

	virtual void add_sample(const Sample &sample, const RadianceSample &radiance) //TODO
	{
	}
	virtual void add_samples(const Sample &sample, const std::list<RadianceSample> &radiances)
	{
		// First the film's samples are found...
		std::list<Vector2> film_samples;
		std::list<Real> temporal_samples;
		Film::get_film_samples(sample.position, film_samples);
		// We iterate over all film samples
		std::list<Vector2>::iterator s;
		std::list<Real>::iterator t;
		for( s = film_samples.begin(); s != film_samples.end(); s++ )
		{
			int x = static_cast<int>((*s)[0]);
			int y = static_cast<int>((*s)[1]);
			
			//Check if new scanline is needed...
			if( y > m_last_available_slice)
			{
				//... if so, get it!
				//	First store the unused scanline...
				fix_and_store_slice(m_first_available_slice);

				//... and remove it from memory.
				delete[] m_available_slices.front();
				m_available_slices.erase(m_available_slices.begin());
				
				// And then, add the new scanline
				m_available_slices.push_back(new Real[3*m_time_resolution*width]());

				if (components & VARIANCE)
				{
					//... and remove it from memory.
					delete[] m_available_slices_variance.front();
					delete[] m_available_slices_nsamples.front();
					m_available_slices_variance.erase(m_available_slices_variance.begin());
					m_available_slices_nsamples.erase(m_available_slices_nsamples.begin());
				
					// And then, add the new scanline
					m_available_slices_variance.push_back(new Real[3*m_time_resolution*width]());
					m_available_slices_nsamples.push_back(new Real[3*m_time_resolution*width]());
				}
				++m_first_available_slice;
				++m_last_available_slice;
			}

			// We first get the weight of the pixel...
			//Filter in the y dimension...
			Real w = filter->evaluate((*s)-sample.position)*sample.weight;
			// ... and store it on weights matrix!
			weight[(y*width+x)] += w;



			// Now, we can iterate over all samples
			std::list<RadianceSample>::const_iterator r;
			
			// Histogram!
			for( r = radiances.begin(); r != radiances.end(); ++r )
			{
				// Get time_coordinate
				Real pos_sensor = ((camera_unwarp?(r->time - r->distance):r->time) - m_offset)/ m_exposure_time;
				
				//printf("%.3f\t",pos_sensor);
				// If out of the range of the streak sensor, then discard.
				if( !(pos_sensor< m_time_resolution && pos_sensor >= 0)) continue;
				// Else, get the current slice, and the temporal samples in the streak film
				int idx = y-m_first_available_slice;
				
                
				
				get_film_samples(pos_sensor, 2, temporal_samples);

				//Finally, just draw the radiances on the streak film, multiplied by the spatio-temporal filtering weight!
				//Note that the temporal weight doesn't act as a filtering weight, but as a PSF. Thus, is additive.
				for( t = temporal_samples.begin(); t != temporal_samples.end(); ++t)
				{				
					//m_available_slices[idx]->draw_pixel(x,static_cast<int>(*t),r->radiance*w*filter->get_subfilter(2)->evaluate(t-pos_sensor));
					Vector3 rad = r->radiance.to_rgb()*w*time_filter->evaluate((*t)-pos_sensor);
					int offset = (static_cast<int>(*t)+x*m_time_resolution)*3;


					m_available_slices[idx][offset  ] += rad[0];
					m_available_slices[idx][offset+1] += rad[1];
					m_available_slices[idx][offset+2] += rad[2];

					if (components & VARIANCE)
					{
						//ONLINE VARIANCE
						Real delta_rad[3];
						for (int c=0; c < 3; c++)
						{
							Real mean=(m_available_slices_nsamples[idx][offset+c])?(m_available_slices[idx][offset+c]/m_available_slices_nsamples[idx][offset+c]):0; //mean = current_value/nsamples;
							m_available_slices_nsamples[idx][offset+c]++; // nsamples++
							delta_rad[c] = rad[c] - mean;
							mean = mean + delta_rad[c]/m_available_slices_nsamples[idx][offset+c];
							m_available_slices_variance[idx][offset+c] += delta_rad[c]*(rad[c] - mean);
						}

					}
					//m_available_slices[idx]->draw_pixel(x,static_cast<int>(*t),r->radiance*w*filter->get_subfilter(2)->evaluate(t-pos_sensor));
				}

				//And store it into the accumulated image
				draw_pixel(x,y,r->radiance.to_rgb()*w);
			}
		}
	}
	void histogram_density_estimation()
	{

	}
	
		
	virtual void clear()
	{
		if( !m_available_slices.empty() )
		{
			for( int slice = m_first_available_slice; slice <= m_last_available_slice; ++slice )
					fix_and_store_slice(slice);

			m_last_available_slice = -1;

			while(!m_available_slices.empty())
			{
				if( m_available_slices.back() )
					delete m_available_slices.back();

				m_available_slices.pop_back();
			}
		}
	}
    Real get_time_length() const 
    {return static_cast<float>(m_time_resolution)*m_exposure_time;}

	virtual Real get_exposure_time ()
	{ return m_exposure_time; }
    
	void set_offset(Real offset)
    //Sets time offset given in picoseconds
	{
		m_offset = offset;
	}
    
    Real get_time_offset() const
    //Returns time offset in picoseconds
	{
		return m_offset;
	}

	void set_streak( int y ) 
	{
		int delta = m_last_available_slice-m_first_available_slice;
		m_first_available_slice = y;
		m_last_available_slice = y+delta;
	}
}; //StreakCameraFilm


#endif //_STREAK_CAMERA_FILM_H_