/* 
* Copyright (C) 2015, Adrian Jarabo (http://giga.cps.unizar.es/~ajarabo/)
*
* Permission is hereby granted, free of charge, to any person obtaining
* a copy of this software and associated documentation files (the "Software"),
* to deal in the Software without restriction, including without limitation
* the rights to use, copy, modify, merge, publish, distribute, sublicense,
* and/or sell copies of the Software, and to permit persons to whom
* the Software is furnished to do so, subject to the following conditions:

* The above copyright notice and this permission notice shall be included
* in all copies or substantial portions of the Software.

* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
* EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
* MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
* IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
* DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
* TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
* OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#ifndef _INTEGRATOR_H_
#define _INTEGRATOR_H_

#include "RayTracing/Intersection.h"
#include "RadianceSample.h"
#include "Color/Spectrum.h"
#include "LinearAlgebra/VectorN.h"
#include <list>

template<int D> class World;

namespace Scattering{
	enum Level{
		NONE = 0,
		SINGLE = 1, 
		MULTIPLE = 2, 
		ALL = SINGLE | MULTIPLE
	};
};

/**	Base virtual class for all integrators. */
template<int D>
class Integrator
{
public:
	World<D> *m_world;
	unsigned int m_incoming_samples;
	Real m_inv_incoming_samples;

	Scattering::Level m_components;
public:
	Integrator(World<D> *w,  unsigned int incoming_samples = 1)
		:m_world(w),m_incoming_samples(incoming_samples), m_inv_incoming_samples(1./static_cast<Real>(m_incoming_samples)),
		m_components(Scattering::ALL){}
	~Integrator(){}

	void set_scattering_components(const Scattering::Level &components)
	{	m_components = components; }

	/** Preprocess function for the integrator. The default behaviour 
		is doing nothing. Other integrators (e.g. these based on 
		particle tracing) would re-implement this function.	*/
	virtual void preprocess(){}
	
	/**	Operator that integrates the radiance incoming to the
		Intersection it. Note that these functions are not thought to
		model volumetric scattering; only surface radiance integration.	
		NOTE: Legacy code; this function is deprecated. */
	virtual Spectrum operator()(const Intersection<D> &it) const = 0;
	
	/**	Operator that integrates the radiance incoming to the
		Intersection it, similar to operator(it), but taking
		into account different time samples. This more addecuate
		for transient rendering.	
		NOTE: Legacy code; this function is deprecated. */
	virtual void operator()(const Intersection<D> &it, std::list<RadianceSample> &samples, const Real delta_time = 0., const unsigned int nb_time_samples = 1, const Real offset_time = 0.) const {};
	
	/**	Operator that integrates the radiance incoming from Ray r.
		This function includes the light incoming from the environment,
		and also the light scattered by the medium through the ray.	*/	
	virtual Spectrum operator()(const Ray<D> &r)const{return Spectrum(0.);}// = 0;
	
	/**	Operator that integrates the radiance incoming to
		r.origin(), from the direction r.direction(). This is similar 
		to operator(r), but taking into account different time samples. 
		This more addecuate for transient rendering.	*/	
	virtual void operator()(const Ray<D> &r, std::list<RadianceSample> &samples, const Real delta_time = 0., const unsigned int nb_time_samples = 1 ) const{}// = 0;
}; //Integrator

typedef Integrator<2> Intergrator2D;
typedef Integrator<3> Intergrator3D;

#endif //_INTEGRATOR_H_