/* 
* Copyright (C) 2015, Adrian Jarabo (http://giga.cps.unizar.es/~ajarabo/)
*
* Permission is hereby granted, free of charge, to any person obtaining
* a copy of this software and associated documentation files (the "Software"),
* to deal in the Software without restriction, including without limitation
* the rights to use, copy, modify, merge, publish, distribute, sublicense,
* and/or sell copies of the Software, and to permit persons to whom
* the Software is furnished to do so, subject to the following conditions:

* The above copyright notice and this permission notice shall be included
* in all copies or substantial portions of the Software.

* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
* EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
* MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
* IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
* DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
* TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
* OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#ifndef _COSINE_LIGHT_SOURCE_H
#define _COSINE_LIGHT_SOURCE_H

#include "Utils/globals.h"
#include "LightSource/PointLightSource.h"
#include "RayTracing/World.h"
#include <math.h>


/** CosineLightSource class represents a point light emiting 
	light with a cosine distribution. */
template<int D>
class CosineLightSource: public PointLightSource<D>
{
protected:
	VectorN<D> direction;
public:
	CosineLightSource( World<D>* _world, const VectorN<D>& pos, const VectorN<D>& dir, const Spectrum& ints, Real time = 0.): 
		PointLightSource<D>(_world, pos, ints,time), direction(dir) {}
	CosineLightSource( World<D>* _world, const Intersection<D> &it, const Spectrum& ints ): 
		PointLightSource<D>(_world, it.get_position(), 
						ints*it.material()->get_absorption(it.get_uv())*dot_clamped(it.get_normal(), -it.get_ray().get_direction()), 
						it.get_ray().get_parameter()/it.get_ray().get_refraction_index()), direction(it.get_normal()) {}

	CosineLightSource( World<D>* _world, const Intersection<D> &it, const Spectrum& ints, Real time ):
		PointLightSource<D>(_world, it.get_position(), 
				ints*it.material()->get_absorption(it.get_uv())*dot_clamped(it.get_normal(), -it.get_ray().get_direction()),
				time), direction(it.get_normal()) {}
  
	Spectrum get_incoming_light(const VectorN<D> &point_lighted) const;
	bool sample( const Intersection<D> &it, LightSample<D> &light_sample, Real &pdf )const;
	void sample( LightSample<D> &light_sample, Real &pdf )const;
	bool sample( const VectorN<D> &p, LightSample<D> &light_sample, Real &pdf )const;

}; //CosineLightSource

template<>
Spectrum CosineLightSource<2>::get_incoming_light(const VectorN<2> &point_lighted) const
{
	VectorN<2> dir = point_lighted-position;
	Real inv_distance = 1./dir.length();
	dir *= inv_distance;
	return intensities*inv_distance*dot_clamped(direction, dir);
}

template<>
Spectrum CosineLightSource<3>::get_incoming_light(const VectorN<3> &point_lighted) const
{
	VectorN<3> dir = point_lighted-position;
	Real distance2 = dir.length2(); Real distance = sqrt(distance2);
	dir /= distance;
	return intensities*(1.f/distance2)*dot_clamped(direction, dir);
}	

template<>
bool CosineLightSource<2>::sample( const Intersection<2> &it, LightSample<2> &light_sample, Real &pdf )const
{
	light_sample.pos = position;
	light_sample.dir = it.get_position()-position;
	light_sample.dist = light_sample.dir.length();
	light_sample.dir /= light_sample.dist;
	light_sample.irradiance = 0.;
	light_sample.instant = time;
	
	pdf = 1.;
	
	Real dot_light = dot(light_sample.dir, direction);
	
	if ( it.get_normal().dot(-light_sample.dir) > 0. && dot_light>0. && is_visible(it.get_position()))
	{
		light_sample.irradiance = intensities / light_sample.dist *dot_light;
		return true;
	}
	
	return false;
}

template<>
bool CosineLightSource<2>::sample( const VectorN<2> &p, LightSample<2> &light_sample, Real &pdf )const
{
	light_sample.pos = position;
	light_sample.dir = p-position;
	light_sample.dist = light_sample.dir.length();
	light_sample.dir /= light_sample.dist;
	light_sample.irradiance = 0.;
	light_sample.instant = time;
	
	pdf = 1.;
	
	Real dot_light = dot(light_sample.dir, direction);
	
	if (dot_light>0. && is_visible(p))
	{
		light_sample.irradiance = intensities / light_sample.dist *dot_light;
		return true;
	}
	
	return false;
}

template<>
bool CosineLightSource<3>::sample( const Intersection<3> &it, LightSample<3> &light_sample, Real &pdf )const
{
	light_sample.pos = position;
	light_sample.dir = it.get_position()-position;
	light_sample.dist = light_sample.dir.length();
	light_sample.dir /= light_sample.dist;
	light_sample.irradiance = 0.;
	light_sample.instant = time;

	pdf = 1.;
	Real dot_light = dot(light_sample.dir, direction);
	
	if( it.get_normal().dot(-light_sample.dir) > 0. && dot_light > 0. && is_visible(it.get_position()))
	{
		light_sample.irradiance = intensities / (light_sample.dist*light_sample.dist)*dot_light;
		return true;
	}

	return false;
}

template<>
bool CosineLightSource<3>::sample( const VectorN<3> &p, LightSample<3> &light_sample, Real &pdf )const
{
	light_sample.pos = position;
	light_sample.dir = p-position;
	light_sample.dist = light_sample.dir.length();
	light_sample.dir /= light_sample.dist;
	light_sample.irradiance = 0.;
	light_sample.instant = time;
	
	pdf = 1.;
	Real dot_light = dot(light_sample.dir, direction);
	
	if( dot_light > 0. && is_visible(p))
	{
		light_sample.irradiance = intensities / (light_sample.dist*light_sample.dist)*dot_light;
		return true;
	}
	
	return false;
}

template<>
void CosineLightSource<2>::sample( LightSample<2> &light_sample, Real &pdf )const
{
	light_sample.pos = position;
	light_sample.dist = 0.;
	light_sample.irradiance = intensities;
	light_sample.instant = time;
	
	// Sample semicircle
	Real epsilon2 = static_cast<Real>(rand())/static_cast<Real>(RAND_MAX);
	
	Real cos_theta, sin_theta;
	
	cos_theta = sqrtf(epsilon2); //sample right or left of the semicircle
	sin_theta = ((rand() % 2 == 1)? 1: -1)*sqrtf(1.-epsilon2);
	
	light_sample.dir = Vector2(sin_theta, cos_theta).transform_matrix_to(Vector2(0,1), direction);
	
	pdf = 0.5*cos_theta; //1./2. (cosine-weighted semicircle)
}
template<> 
void CosineLightSource<3>::sample( LightSample<3> &light_sample, Real &pdf )const
{
	light_sample.pos = position;
	light_sample.dist = 0.;
	light_sample.irradiance = intensities;
	light_sample.instant = time;

	// The sampling pipeline should be improved, but for now it's more than enough...
	// Sample sphere 
	Real epsilon1 = static_cast<Real>(rand())/static_cast<Real>(RAND_MAX);
	while (epsilon1 < 0.) 
		epsilon1 = static_cast<Real>(rand())/static_cast<Real>(RAND_MAX);

	Real epsilon2 = static_cast<Real>(rand())/static_cast<Real>(RAND_MAX);
	while (epsilon2 < 1e-5) 
		epsilon2 = static_cast<Real>(rand())/static_cast<Real>(RAND_MAX);

	Real phi = 2*M_PI*epsilon1;
	Real cos_thita = sqrt(epsilon2);//2*epsilon2-1.;
	Real sin_thita = sqrt(1.-cos_thita*cos_thita);

		light_sample.dir = Vector3(cos(phi)*sin_thita, cos_thita, sin(phi)*sin_thita).transform_matrix_to(Vector3(0,1,0), direction);

		pdf = M_1_PI*cos_thita; //1./PI (cosine-weighted hemisphere)
	}

#endif //_COSINE_LIGHT_SOURCE_H


