/* 
* Copyright (C) 2015, Adrian Jarabo (http://giga.cps.unizar.es/~ajarabo/)
*
* Permission is hereby granted, free of charge, to any person obtaining
* a copy of this software and associated documentation files (the "Software"),
* to deal in the Software without restriction, including without limitation
* the rights to use, copy, modify, merge, publish, distribute, sublicense,
* and/or sell copies of the Software, and to permit persons to whom
* the Software is furnished to do so, subject to the following conditions:

* The above copyright notice and this permission notice shall be included
* in all copies or substantial portions of the Software.

* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
* EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
* MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
* IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
* DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
* TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
* OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#ifndef _DIRECTIONAL_LIGHT_SOURCE_H_
#define _DIRECTIONAL_LIGHT_SOURCE_H_

#include "LightSource/LightSource.h"
#include "LinearAlgebra/Vector3.h"
#include "RayTracing/World.h"
#include "RayTracing/Ray.h"
#include "RayTracing/Intersection.h"

/** DirectionalLightSource class represents a directional light sending light equally
 in a single direction without attenuation. */
template<int D>
class DirectionalLightSource : public LightSource<D>
{
protected:
	VectorN<D> direction;
	
public:
	DirectionalLightSource(World<D>* _world, const VectorN<D>& dir, const Spectrum& ints, Real time = 0.): 
	LightSource<D>(_world,ints,time),	direction(normalize(dir)) {}
	
	VectorN<D> get_direction() const { return direction; }
	VectorN<D> get_incoming_direction(const VectorN<D> &point_lighted) const
	{
		return direction;
	}
	bool is_visible(const VectorN<D> &point_lighted) const
	{
		VectorN<D> dir = direction;
		Ray<D> ray(point_lighted, -dir, false); ray.shift();
		
		return( !this->world->intersects(ray)) ;
	}
	virtual Spectrum get_incoming_light(const VectorN<D> &point_lighted) const;
	
	virtual bool sample( const Intersection<D> &it, LightSample<D> &light_sample, Real &pdf )const;
	virtual bool sample( const VectorN<D> &p, LightSample<D> &light_sample, Real &pdf )const;
	
	virtual void sample( LightSample<D> &light_sample, Real &pdf )const;
}; //DirectionalLightSource


template<int D>
Spectrum DirectionalLightSource<D>::get_incoming_light(const VectorN<D> &point_lighted) const
{
	return is_visible(point_lighted)?intensities:Spectrum(0.);
}

template<int D>
bool DirectionalLightSource<D>::sample( const Intersection<D> &it, LightSample<D> &light_sample, Real &pdf )const
{
	light_sample.pos = VectorN<D>(0.);
	light_sample.dir = direction;
	light_sample.dist = 0.;
	light_sample.irradiance = 0.;
	light_sample.instant = time;
	
	pdf = 1.;
	
	if (it.get_normal().dot(-light_sample.dir) > 0. && is_visible(it.get_position()))
	{
		light_sample.irradiance = intensities;
		return true;
	}
	
	return false;
}

template<int D>
bool DirectionalLightSource<D>::sample( const VectorN<D> &p, LightSample<D> &light_sample, Real &pdf )const
{
	light_sample.pos = VectorN<D>(0.);
	light_sample.dir = direction;
	light_sample.dist = 0.;
	light_sample.irradiance = 0.;
	light_sample.instant = time;
	
	pdf = 1.;
	
	if (is_visible(p))
	{
		light_sample.irradiance = intensities;
		return true;
	}
	
	return false;
}

template<int D>
void DirectionalLightSource<D>::sample( LightSample<D> &light_sample, Real &pdf )const
{
	light_sample.pos = VectorN<D>(0.);
	light_sample.dist = 0.;
	light_sample.irradiance = intensities;
	light_sample.instant = time;
	light_sample.dir = direction;
	
	pdf = 1.;
}

#endif //_DIRECTIONAL_LIGHT_SOURCE_H_