/* 
* Copyright (C) 2015, Adrian Jarabo (http://giga.cps.unizar.es/~ajarabo/)
*
* Permission is hereby granted, free of charge, to any person obtaining
* a copy of this software and associated documentation files (the "Software"),
* to deal in the Software without restriction, including without limitation
* the rights to use, copy, modify, merge, publish, distribute, sublicense,
* and/or sell copies of the Software, and to permit persons to whom
* the Software is furnished to do so, subject to the following conditions:

* The above copyright notice and this permission notice shall be included
* in all copies or substantial portions of the Software.

* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
* EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
* MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
* IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
* DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
* TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
* OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#ifndef _POINT_LIGHT_SOURCE_H_
#define _POINT_LIGHT_SOURCE_H_

#include "LightSource/LightSource.h"
#include "LinearAlgebra/Vector3.h"
#include "RayTracing/World.h"
#include "RayTracing/Ray.h"
#include "RayTracing/Intersection.h"
#include "Utils/RandomNumbers.h"

/** PointLightSource class represents a single point sending light equally
		in all directions. */
template<int D>
class PointLightSource : public LightSource<D>
{
protected:
	VectorN<D> position;

public:
	PointLightSource(World<D>* _world, const VectorN<D>& pos, const Spectrum& ints, Real time = 0.): 
		LightSource<D>(_world,ints,time),	position(pos) {}
  
	VectorN<D> get_position() const { return position; }
	VectorN<D> get_incoming_direction(const VectorN<D> &point_lighted) const
	{
		return normalize(point_lighted-position);
	}
	bool is_visible(const VectorN<D> &point_lighted) const
	{
		VectorN<D> dir = point_lighted-position;
		Real l = dir.length(); dir /= l;
		Ray<D> ray(position, dir, false); ray.shift();
		return( !this->world->intersects(ray, l, _SIGMA_VISIBILITY_)) ;
		
				
		
		Intersection<D> it;
		this->world->first_intersection(ray,it,l+100);

		printf("Distance: %f vs %f [%d]", l, ray.get_parameter(), ray.did_hit());
		if( ray.did_hit() && ray.get_parameter() < l )
			return false;

		return true;
	}
	virtual Spectrum get_incoming_light(const VectorN<D> &point_lighted) const;

	virtual bool sample( const Intersection<D> &it, LightSample<D> &light_sample, Real &pdf )const;

	virtual void sample( LightSample<D> &light_sample, Real &pdf )const;
	
	virtual bool sample( const VectorN<D> &p, LightSample<D> &light_sample, Real &pdf )const;
}; //PointLightSource

//Specializations for 2D and 3D
template<>
Spectrum PointLightSource<2>::get_incoming_light(const VectorN<2> &point_lighted) const
{
	Real distance = (point_lighted-position).length();
	return this->intensities*(1.f/distance);
}
template<>
Spectrum PointLightSource<3>::get_incoming_light(const VectorN<3> &point_lighted) const
{
	Real distance2 = (point_lighted-position).length2();
	return this->intensities*(1.f/distance2);
}	


template<>
bool PointLightSource<2>::sample( const Intersection<2> &it, LightSample<2> &light_sample, Real &pdf )const
{
	light_sample.pos = position;
	light_sample.dir = it.get_position()-position;
	light_sample.dist = light_sample.dir.length();
	light_sample.dir /= light_sample.dist;
	light_sample.irradiance = 0.;
	light_sample.instant = time;
	
	pdf = 1.;
	
	if (it.get_normal().dot(-light_sample.dir) > 0. && is_visible(it.get_position()))
	{
		light_sample.irradiance = intensities / light_sample.dist;
		return true;
	}
	
	return false;
}

template<>
bool PointLightSource<2>::sample( const VectorN<2> &p, LightSample<2> &light_sample, Real &pdf )const
{
	light_sample.pos = position;
	light_sample.dir = p-position;
	light_sample.dist = light_sample.dir.length();
	light_sample.dir /= light_sample.dist;
	light_sample.irradiance = 0.;
	light_sample.instant = time;
	
	pdf = 1.;
	
	if (is_visible(p))
	{
		light_sample.irradiance = intensities / light_sample.dist;
		return true;
	}
	
	return false;
}

template<>
bool PointLightSource<3>::sample( const Intersection<3> &it, LightSample<3> &light_sample, Real &pdf )const
{
	light_sample.pos = position;
	light_sample.dir = it.get_position()-position;
	light_sample.dist = light_sample.dir.length();
	light_sample.dir /= light_sample.dist;
	light_sample.irradiance = 0.;
	light_sample.instant = time;
	
	pdf = 1.;
	
	if( it.get_normal().dot(-light_sample.dir) > 0. && is_visible(it.get_position()))
	{
		light_sample.irradiance = intensities / (light_sample.dist*light_sample.dist);
		return true;
	}
	
	return false;
}

template<>
bool PointLightSource<3>::sample( const VectorN<3> &p, LightSample<3> &light_sample, Real &pdf )const
{
	light_sample.pos = position;
	light_sample.dir = p-position;
	light_sample.dist = light_sample.dir.length();
	light_sample.dir /= light_sample.dist;
	light_sample.irradiance = 0.;
	light_sample.instant = time;
	
	pdf = 1.;
	
	if(is_visible(p))
	{
		light_sample.irradiance = intensities / (light_sample.dist*light_sample.dist);
		return true;
	}
	
	return false;
}

template<>
void PointLightSource<2>::sample( LightSample<2> &light_sample, Real &pdf )const
{
	light_sample.pos = position;
	light_sample.dist = 0.;
	light_sample.irradiance = intensities;
	light_sample.instant = time;
	//Real epsilon1 = RNG::StdRandom::get_real();
	Real epsilon1 = static_cast<Real>(rand())/(static_cast<Real>(RAND_MAX)+1.0);
	while (epsilon1 < 0.) {
		epsilon1 = static_cast<Real>(rand())/(static_cast<Real>(RAND_MAX)+1.0);
	}
	Real phi = 2.*M_PI*epsilon1;	
	
	light_sample.dir = Vector2(cos(phi), sin(phi));
	
	pdf = M_1_PI*.5; //1/2PI (1/circumference)
}

template<>
void PointLightSource<3>::sample( LightSample<3> &light_sample, Real &pdf )const
{
	light_sample.pos = position;
	light_sample.dist = 0.;
	light_sample.irradiance = intensities;
	light_sample.instant = time;

	// The sampling pipeline should be improved, but for now it's more than enough...
	// Sample sphere 
	Real epsilon1 = RNG::StdRandom::get_real();//static_cast<Real>(rand())/static_cast<Real>(RAND_MAX);
	Real epsilon2 = RNG::StdRandom::get_real();;//static_cast<Real>(rand())/static_cast<Real>(RAND_MAX);

	Real phi = 2*M_PI*epsilon1;
	Real cos_thita = 2*epsilon2-1.;
	Real sin_thita = sqrt(1.-cos_thita*cos_thita);

	light_sample.dir = Vector3(cos(phi)*sin_thita, cos_thita, sin(phi)*sin_thita);

	pdf = M_1_PI*.25; //1./(4*PI)
}


#endif //_POINT_LIGHT_SOURCE_H_