/* 
* Copyright (C) 2015, Adrian Jarabo (http://giga.cps.unizar.es/~ajarabo/)
*
* Permission is hereby granted, free of charge, to any person obtaining
* a copy of this software and associated documentation files (the "Software"),
* to deal in the Software without restriction, including without limitation
* the rights to use, copy, modify, merge, publish, distribute, sublicense,
* and/or sell copies of the Software, and to permit persons to whom
* the Software is furnished to do so, subject to the following conditions:

* The above copyright notice and this permission notice shall be included
* in all copies or substantial portions of the Software.

* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
* EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
* MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
* IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
* DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
* TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
* OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#ifndef _SPOT_LIGHT_SOURCE_H
#define _SPOT_LIGHT_SOURCE_H

#include "Utils/globals.h"
#include "LightSource/PointLightSource.h"
#include "RayTracing/World.h"
#include <math.h>


/** SpotLightSource class represents a spot point light emiting. 
	TODO: Comment this properly...*/
template<int D>
class SpotLightSource: public PointLightSource<D>
{
protected:
	Real m_angle;
	Real m_cos_angle;
	Real m_p;
	VectorN<D> direction;
public:
	SpotLightSource( World<D>* _world, const VectorN<D>& pos, const VectorN<D>& target, const Real &angle, const Spectrum& ints, Real time = 0.): 
	  PointLightSource<D>(_world, pos, ints, time), direction(normalize(target-pos)), m_angle(angle*M_PI/360), m_cos_angle((angle>180.)?0:cos(m_angle)), 
		  m_p((D==2)?(m_angle):(2*M_PI*(1-m_cos_angle))) {}
  
	Spectrum get_incoming_light(const VectorN<D> &point_lighted) const;
	bool sample( const Intersection<D> &it, LightSample<D> &light_sample, Real &pdf )const;
	void sample( LightSample<D> &light_sample, Real &pdf )const;
	bool sample( const VectorN<D> &p, LightSample<D> &light_sample, Real &pdf )const;

}; //SpotLightSource

template<>
Spectrum SpotLightSource<2>::get_incoming_light(const VectorN<2> &point_lighted) const
{
	VectorN<2> dir = point_lighted-position;
	Real inv_distance = 1./dir.length();
	dir *= inv_distance;
	return intensities*inv_distance*((dot(direction,dir)<m_cos_angle)?0.:1.);
}

template<>
Spectrum SpotLightSource<3>::get_incoming_light(const VectorN<3> &point_lighted) const
{
	VectorN<3> dir = point_lighted-position;
	Real distance2 = dir.length2(); Real distance = sqrt(distance2);
	dir /= distance;
	return intensities*(1.f/distance2)*((dot(direction,dir)<m_cos_angle)?0.:1.);
}	

template<>
bool SpotLightSource<2>::sample( const Intersection<2> &it, LightSample<2> &light_sample, Real &pdf )const
{
	light_sample.pos = position;
	light_sample.dir = it.get_position()-position;
	light_sample.dist = light_sample.dir.length();
	light_sample.dir /= light_sample.dist;
	light_sample.irradiance = 0.;
	light_sample.instant = time;
	
	pdf = 1.;
	
	Real dot_light = dot(light_sample.dir, direction);
	
	if ( it.get_normal().dot(-light_sample.dir) > 0. && dot_light>=m_cos_angle && is_visible(it.get_position()))
	{
		light_sample.irradiance = intensities / light_sample.dist;
		return true;
	}
	
	return false;
}

template<>
bool SpotLightSource<2>::sample( const VectorN<2> &p, LightSample<2> &light_sample, Real &pdf )const
{
	light_sample.pos = position;
	light_sample.dir = p-position;
	light_sample.dist = light_sample.dir.length();
	light_sample.dir /= light_sample.dist;
	light_sample.irradiance = 0.;
	light_sample.instant = time;
	
	pdf = 1.;
	
	Real dot_light = dot(light_sample.dir, direction);
	
	if (dot_light>=m_cos_angle && is_visible(p))
	{
		light_sample.irradiance = intensities / light_sample.dist;
		return true;
	}
	
	return false;
}

template<>
bool SpotLightSource<3>::sample( const Intersection<3> &it, LightSample<3> &light_sample, Real &pdf )const
{
	light_sample.pos = position;
	light_sample.dir = it.get_position()-position;
	light_sample.dist = light_sample.dir.length();
	light_sample.dir /= light_sample.dist;
	light_sample.irradiance = 0.;
	light_sample.instant = time;

	pdf = 1.;
	Real dot_light = dot(light_sample.dir, direction);
	
	//printf("Sampling light source. IRR: %f\n", intensities[0]);
	if( it.get_normal().dot(-light_sample.dir) > 0. && dot_light>=m_cos_angle && is_visible(it.get_position()))
	{
		light_sample.irradiance = intensities / (light_sample.dist*light_sample.dist);
		return true;
	}

	return false;
}

template<>
bool SpotLightSource<3>::sample( const VectorN<3> &p, LightSample<3> &light_sample, Real &pdf )const
{
	light_sample.pos = position;
	light_sample.dir = p-position;
	light_sample.dist = light_sample.dir.length();
	light_sample.dir /= light_sample.dist;
	light_sample.irradiance = 0.;
	light_sample.instant = time;
	
	pdf = 1.;
	Real dot_light = dot(light_sample.dir, direction);
	
	if( dot_light>=m_cos_angle && is_visible(p))
	{
		light_sample.irradiance = intensities / (light_sample.dist*light_sample.dist);
		return true;
	}
	
	return false;
}

template<>
void SpotLightSource<2>::sample( LightSample<2> &light_sample, Real &pdf )const
{
	light_sample.pos = position;
	light_sample.dist = 0.;
	light_sample.irradiance = intensities;
	light_sample.instant = time;
	
	// Sample semicircle
	Real epsilon2;
	Real cos_theta(-1.), sin_theta;
	
	while (cos_theta < m_cos_angle)
	{
		epsilon2 = static_cast<Real>(rand())/static_cast<Real>(RAND_MAX);
		cos_theta = sqrtf(epsilon2); 
	}

	//sample right or left of the semicircle
	sin_theta = ((rand() % 2 == 1)? 1: -1)*sqrtf(1.-epsilon2);
	
	light_sample.dir = Vector2(sin_theta, cos_theta).transform_matrix_to(Vector2(0,1), direction);
	
	pdf = m_p; //1./2. (cosine-weighted semicircle)
}
template<> 
void SpotLightSource<3>::sample( LightSample<3> &light_sample, Real &pdf )const
{
	light_sample.pos = position;
	light_sample.dist = 0.;
	light_sample.irradiance = intensities;
	light_sample.instant = time;

	Real cos_thita(-1.), sin_thita, epsilon2;


	// The sampling pipeline should be improved, but for now it's more than enough...
	// Sample sphere 
	Real epsilon1 = static_cast<Real>(rand())/static_cast<Real>(RAND_MAX);
	while (epsilon1 < 0.) 
		epsilon1 = static_cast<Real>(rand())/static_cast<Real>(RAND_MAX);
	Real phi = 2*M_PI*epsilon1;


	while (cos_thita < m_cos_angle)
	{
		epsilon2 = static_cast<Real>(rand())/static_cast<Real>(RAND_MAX);
		while (epsilon2 < 1e-5) 
			epsilon2 = static_cast<Real>(rand())/static_cast<Real>(RAND_MAX);

		cos_thita = sqrt(epsilon2);//2*epsilon2-1.;
		sin_thita = sqrt(1.-cos_thita*cos_thita);
	}

	light_sample.dir = Vector3(cos(phi)*sin_thita, cos_thita, sin(phi)*sin_thita).transform_matrix_to(Vector3(0,1,0), direction);

	pdf = m_p; //1./PI (cosine-weighted hemisphere)
}

#endif //_SPOT_LIGHT_SOURCE_H


