/* 
* Copyright (C) 2015, Adrian Jarabo (http://giga.cps.unizar.es/~ajarabo/)
*
* Permission is hereby granted, free of charge, to any person obtaining
* a copy of this software and associated documentation files (the "Software"),
* to deal in the Software without restriction, including without limitation
* the rights to use, copy, modify, merge, publish, distribute, sublicense,
* and/or sell copies of the Software, and to permit persons to whom
* the Software is furnished to do so, subject to the following conditions:

* The above copyright notice and this permission notice shall be included
* in all copies or substantial portions of the Software.

* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
* EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
* MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
* IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
* DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
* TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
* OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#ifndef _VECTORN_H_
#define _VECTORN_H_

#include "Utils/globals.h"
#include <math.h>
#include <algorithm>

template<int N>
class VectorN
{
public:
	Real m_data[N];
public:
	// Constructors
	VectorN(const Real f= 0.f)
	{
		for(int i=0; i<N; ++i)
			m_data[i] = f;

	}
	/** The following constructors are very useful in the case
		of Vecotor<2> and Vector<3>, which are actually the most
		common types of vectors in rendering					*/
	VectorN(const Real x, const Real y)
	{
		if( N == 2 )
		{
			m_data[0]=x;
			m_data[1]=y;
		}
		else
		{
			m_data[0] = x;
			if(N>1)
			{	
				m_data[1] = y;
				for( int i=2; i<N; ++i ) m_data[i] = 0.;
			}
		}
	}
	VectorN(const Real x, const Real y, const Real z)
	{
		if( N == 3 )
		{
			m_data[0]=x;
			m_data[1]=y;
			m_data[2]=z;
		}
		else
		{
			m_data[0] = x;
			if(N>1)
			{	
				m_data[1] = y;
				if( N>2 )
				{
					m_data[2] = z;
					for( int i=2; i<N; ++i ) m_data[i] = 0.;
				}
			}
		}
	}
	VectorN(const VectorN<N> &v)
	{
		for(int i=0; i<N; ++i)
			m_data[i] = v.m_data[i];

	}
	// Arithmetic operators
	VectorN<N> operator-()const
	{
		VectorN<N> nspectrum;
		for( int i=0; i<N; ++i)
			nspectrum[i] = -m_data[i];

		return nspectrum;
	}
	VectorN<N> operator+(const VectorN<N>& v) const
	{
		VectorN<N> nspectrum;
		for( int i=0; i<N; ++i)
			nspectrum[i] = m_data[i]+v.m_data[i];

		return nspectrum;
	}
	VectorN<N> operator-(const VectorN<N>& v) const
	{
		VectorN<N> nspectrum;
		for( int i=0; i<N; ++i)
			nspectrum[i] = m_data[i]-v.m_data[i];

		return nspectrum;
	}
	VectorN<N> operator*(const Real f) const
	{
		VectorN<N> nspectrum;
		for( int i=0; i<N; ++i)
			nspectrum[i] = m_data[i]*f;

		return nspectrum;
	}
	VectorN<N> operator*(const VectorN<N>& v) const
	{
		VectorN<N> nspectrum;
		for( int i=0; i<N; ++i)
			nspectrum[i] = m_data[i]*v.m_data[i];

		return nspectrum;
	}
	VectorN<N> operator/(const Real f) const
	{
		VectorN<N> nspectrum;
		Real inv_f = 1./f;
		for( int i=0; i<N; ++i)
			nspectrum[i] = m_data[i]*inv_f;

		return nspectrum;
	}
	VectorN<N> operator/(const VectorN<N>& v) const
	{
		VectorN<N> nspectrum;
		for( int i=0; i<N; ++i)
			nspectrum[i] = m_data[i]/v.m_data[i];

		return nspectrum;
	}

	const VectorN<N> &operator+=(const VectorN<N>& v)
	{
		for( int i=0; i<N; ++i)
			m_data[i]+=v.m_data[i];

		return *this;
	}
	const VectorN<N> &operator-=(const VectorN<N>& v)
	{
		for( int i=0; i<N; ++i)
			m_data[i]-=v.m_data[i];

		return *this;
	}
	const VectorN<N> &operator*=(const VectorN<N>& v)
	{
		for( int i=0; i<N; ++i)
			m_data[i]*=v.m_data[i];

		return *this;
	}
	const VectorN<N> &operator/=(const VectorN<N>& v)
	{
		for( int i=0; i<N; ++i)
			m_data[i]/=v.m_data[i];

		return *this;
	}

	const VectorN<N> &operator*=(const Real f)
	{
		for( int i=0; i<N; ++i)
			m_data[i]*=f;

		return *this;
	}
	const VectorN<N> &operator/=(const Real f)
	{
		Real inv_f = 1./f;
		for( int i=0; i<N; ++i)
			m_data[i]*=inv_f;

		return *this;
	}

	// Assignation operators
	const VectorN<N> &operator=(const VectorN<N>& v)
	{
		for( int i=0; i<N; ++i)
			m_data[i]=v.m_data[i];

		return *this;
	}

	const VectorN<N> &operator=(const Real f)
	{
		for( int i=0; i<N; ++i)
			m_data[i]=f;

		return *this;
	}

	// Access functions. Define _SAFE_CHECK_ at 'Utils/globals.h'
	// to check if the dimension is in the range [0,2]
	inline Real operator[](const int dimension) const
	{
		#ifdef _SAFE_CHECK_		
			if (dimension > N-1 || dimension < 0) throw("Out-of-range");
		#endif	
		return m_data[dimension];	
	}
	inline Real &operator[](const int dimension)
	{
		#ifdef _SAFE_CHECK_		
			if (dimension > N-1 || dimension < 0) throw("Out-of-range");
		#endif	
		return m_data[dimension];	
	}
	// Functions
	inline Real total()const
	{
		Real value = 0.;
		for( int i=0; i<N; ++i )
			value += m_data[i];

		return value;
	}
	inline Real avg()const
	{
		return total()/static_cast<Real>(N);
	}
	inline bool isZero(void) const
	{
		for( int i=0; i<N; ++i )
			if( m_data[i] != 0. )
				return false;

		return true;
	}
	inline Real length() const
	{
		return sqrtf(length2());
	}
	inline Real length2() const
	{
		Real value = 0.;
		for( int i=0; i<N; ++i )
			value += (m_data[i]*m_data[i]);
		
		return value;
	}
	unsigned int dominant() const
	{
		Real dominant_value = fabs(m_data[0]);
		int dominant_idx = 0;
		
		for( int i=1; i<N; ++i)
		{
			if( fabs(m_data[i]) > dominant_value )
			{
				dominant_value = fabs(m_data[i]);
				dominant_idx = i;
			}
		}
		return dominant_idx;
	}

	inline VectorN<N> normalize(void)
	{
		return ((*this) /= length());
	}

	inline Real min() const
	{
		Real min_v = m_data[0];
		for(int i=1; i<N; ++i )
			min_v = (m_data[i]<min_v)?m_data[i]:min_v;
		
		return min_v;
	}

	inline VectorN<N> min(const VectorN<N> &v) const
	{
		VectorN<N> min_v;
		for(int i=0; i<N; ++i )
			min_v[i] = (m_data[i]<v.m_data[i])?m_data[i]:v.m_data[i];
		
		return min_v;
	}
	
	inline Real max() const
	{
		Real max_v = m_data[0];
		for(int i=1; i<N; ++i )
			max_v = (m_data[i]>max_v)?m_data[i]:max_v;
		
		return max_v;
	}

	inline VectorN<N> max(const VectorN<N> &v) const
	{
		VectorN<N> max_v;
		for(int i=0; i<N; ++i )
			max_v[i] = (m_data[i]>v.m_data[i])?m_data[i]:v.m_data[i];
		
		return max_v;
	}
	inline Real dot(const VectorN<N> &v) const
	{
		Real value = 0.;
		for(int i=0; i<N; ++i )
			value += m_data[i] * v.m_data[i];
		
		return value;
	}
	inline Real dot_clamped(const VectorN<N> &v) const
	{
		Real result = dot(v);
		return (result > 0.)?result:0.; 
	}
	inline Real dot_abs(const VectorN<N> &v) const
	{
		return fabs(dot(v));
	}
	inline VectorN<N> cross(const VectorN<N> &v) const;

	//Function that returns the reflected direction, depending on normal
	VectorN<N> reflect(const VectorN<N> &normal)const
	{	return  ((*this) - 2.f*dot(normal)*normal);		}
	//Function that returns the reflected direction, depending on normal and refractive_index
	VectorN<N> refract(const VectorN<N> &normal, Real eta)const
	{
		/*Real cos_i = -dot(normal);
		Real n = (cos_i<0.)?index_of_refraction:(1.f/index_of_refraction);
		//Real n = 1.f/ior;

		Real cos_o = sqrt(std::max(0.f,1.f-n*n*(1.f-cos_i*cos_i)))*(cos_i>0.?1.:-1.);

		return ((n*(*this)+(n*cos_i-cos_o)*normal));*/


		Real cos_i = dot(normal);
		Real cos_o = sqrtf(std::max(0.f, 1.f-eta*eta*(1.f-cos_i*cos_i)));
		
		//Vector3 wo = ((eta*(*this)+(eta*fabs(cos_i)-cos_o)*(cos_i<0.f?1.f:-1.f)*normal));

		//printf("Eta: %f -> cos_i: %f; cos_o: %f; cos_o': %f\n", eta, cos_i, cos_o, wo.dot(normal));

		return ((eta*(*this)+(eta*fabs(cos_i)-cos_o)*(cos_i<0.f?1.f:-1.f)*normal));
	}

	
	//Functions that change (direction) vector to other coordinate system
	VectorN<N> transform_matrix_to(const VectorN<N>& from, const VectorN<N>& to)const;
	VectorN<N> transform_matrix_to_z(const VectorN<N>& from, const VectorN<N>& from_up)const;
}; //VectorN

/** Template specializations */
// Cross product
template<int N>
VectorN<N> VectorN<N>::cross(const VectorN<N> &v) const
{	throw("No cross product operation defined for vectors of this dimension");	}

template<>
VectorN<3> VectorN<3>::cross(const VectorN<3> &v) const
{
	return VectorN<3>(	(m_data[1] * v.m_data[2] - m_data[2] * v.m_data[1]), 
						(m_data[2] * v.m_data[0] - m_data[0] * v.m_data[2]), 
						(m_data[0] * v.m_data[1] - m_data[1] * v.m_data[0]) );
}

template<>
VectorN<2> VectorN<2>::cross(const VectorN<2> &v) const
{
	return VectorN<2>(m_data[0] * v.m_data[1] - m_data[1] * v.m_data[0], 0.);
}

// transform_matrix_to
template<int N>
VectorN<N> VectorN<N>::transform_matrix_to(const VectorN<N>& from, const VectorN<N>& to) const
{	throw("No \"transform_matrix_to\" operation defined for vectors of this dimension");	}

template<>
VectorN<2> VectorN<2>::transform_matrix_to(const VectorN<2>& from, const VectorN<2>& to) const
{	
	Real mtx[2][2];
	
	Real cos_theta = from.dot(to);
	Real sin_theta = from.cross(to)[0];//sqrt(1.-cos_theta*cos_theta);
		
	VectorN<2> sol;
	
	mtx[0][0] = cos_theta;  // cos    -sin
	mtx[0][1] = -sin_theta; 
	mtx[1][0] = sin_theta;  
	mtx[1][1] = cos_theta;  // sin     cos
	
	for (int i=0;i<2;i++) sol[i]=mtx[i][0]*m_data[0]+mtx[i][1]*m_data[1];

	return sol;
}

template<>
VectorN<3> VectorN<3>::transform_matrix_to(const VectorN<3>& from, const VectorN<3>& to)const
{
	Real mtx[3][3];
	const Real epsilon = 1.e-4;
	VectorN<3> v =from.cross(to);
	Real e = from.dot(to);
	Real f = (e<0.0)?-e:e;
	if (f>1.0-epsilon)
	{
		VectorN<3> u, x;
		Real c1, c2, c3;
		int i, j;

    		x[0] = (from[0] > 0.0)? from[0] : -from[0];
    		x[1] = (from[1] > 0.0)? from[1] : -from[1];
    		x[2] = (from[2] > 0.0)? from[2] : -from[2];
	
		if (x[0] < x[1])
			{      
			if (x[0] < x[2])    { x[0] = 1.0; x[1] = x[2] = 0.0; }
      			else 		    { x[2] = 1.0; x[0] = x[1] = 0.0; }
  			}
		else
		{
  				if (x[1] < x[2])    { x[1] = 1.0; x[0] = x[2] = 0.0; }
  				else                { x[2] = 1.0; x[0] = x[1] = 0.0; }
			}

			u[0] = x[0] - from[0]; u[1] = x[1] - from[1]; u[2] = x[2] - from[2];
			v[0] = x[0] - to[0];   v[1] = x[1] - to[1];   v[2] = x[2] - to[2];

			c1 = 2.0 / u.dot(u);
			c2 = 2.0 / v.dot(v);
			c3 = c1 * c2  * u.dot(v);

			for (i = 0; i < 3; i++) {
  				for (j = 0; j < 3; j++) mtx[i][j] =  -c1*u[i]*u[j] - c2*v[i]*v[j] + c3*v[i]*u[j];
  				mtx[i][i] += 1.0;
		}
	}
	else  /* the most common case, unless "from"="to", or "from"=-"to" */
	{
		/* ...otherwise use this hand optimized version (9 mults less) */
		Real h, hvx, hvz, hvxy, hvxz, hvyz;
		/* h = (1.0 - e)/DOT(v, v); old code */
		h = 1.0/(1.0 + e);      /* optimization by Gottfried Chen */
		hvx = h * v[0];
		hvz = h * v[2];
		hvxy = hvx * v[1];
		hvxz = hvx * v[2];
		hvyz = hvz * v[1];
		mtx[0][0] = e + hvx * v[0];
		mtx[0][1] = hvxy - v[2];
		mtx[0][2] = hvxz + v[1];

		mtx[1][0] = hvxy + v[2];
		mtx[1][1] = e + h * v[1] * v[1];
		mtx[1][2] = hvyz - v[0];

		mtx[2][0] = hvxz - v[1];
		mtx[2][1] = hvyz + v[0];
		mtx[2][2] = e + hvz * v[2];
	  }

	VectorN<3> sol;
	for (int i=0;i<3;i++) sol[i]=mtx[i][0]*m_data[0]+mtx[i][1]*m_data[1]+mtx[i][2]*m_data[2];
	sol.normalize();
	return sol;
}





// transform_matrix_to_z
template<int N>
VectorN<N> VectorN<N>::transform_matrix_to_z(const VectorN<N>& from, const VectorN<N>& from_up)const
{	throw("No \"transform_matrix_to_z\" operation defined for vectors of this dimension");	}
template<>
VectorN<2> VectorN<2>::transform_matrix_to_z(const VectorN<2>& from, const VectorN<2>& from_up) const
{	throw("No \"transform_matrix_to_z\" operation defined for Vector2. At some point it will be implemented...Sorry about that!");	}

template<>
VectorN<3> VectorN<3>::transform_matrix_to_z(const VectorN<3>& from, const VectorN<3>& from_up)const
{
	//Vector3 aux_blah(from[0], 0.,from[2]); aux_blah.normalize();
	VectorN<3> f_up = from_up;
	Real dotup = from.dot(f_up);
	if( fabs(dotup) > .999)
		f_up = VectorN<3>(0.,0.,dotup);

	VectorN<3> right = f_up.cross(from);right.normalize();
	VectorN<3> up = from.cross(right);up.normalize();
	
	VectorN<3> s1(this->dot(right), this->dot(f_up), this->dot(from));
	s1.normalize();

	return s1;
}






/** Non-member operations involving Vectors	*/
template<int N>
inline VectorN<N> normalize(const VectorN<N> &v)
{
	VectorN<N> result(v);
	result.normalize();
	return result;
}
template<int N>
inline VectorN<N> minimum(const VectorN<N> &v1, const VectorN<N> &v2)
{
	return v1.min(v2);
}
template<int N>
inline VectorN<N> maximum(const VectorN<N> &v1, const VectorN<N> &v2)
{
	return v1.max(v2);
}
template<int N>
inline Real dot(const VectorN<N> &v1, const VectorN<N> &v2)
{
	return v1.dot(v2);
}
template<int N>
inline Real dot_clamped(const VectorN<N> &v1, const VectorN<N> &v2)
{
	return v1.dot_clamped(v2);
}
template<int N>
inline Real dot_abs(const VectorN<N> &v1, const VectorN<N> &v2)
{
	return v1.dot_abs(v2);
}
template<int N>
inline VectorN<N> cross(const VectorN<N> &v1, const VectorN<N> &v2)
{
	return v1.cross(v2);
}
template<int N>
inline VectorN<N> operator+(const Real f, const VectorN<N> &v)
{
	return v+f;
}
template<int N>
inline VectorN<N> operator-(const Real f, const VectorN<N> &v)
{
	return VectorN<N>(f)-v;
}
template<int N>
inline VectorN<N> operator*(const Real f, const VectorN<N> &v)
{
	return v*f;
}
template<int N>
inline VectorN<N> operator/(const Real f, const VectorN<N> &v)
{
	return VectorN<N>(f)/v;
}


#endif //_VECTORN_H_