/* 
* Copyright (C) 2015, Adrian Jarabo (http://giga.cps.unizar.es/~ajarabo/)
*
* Permission is hereby granted, free of charge, to any person obtaining
* a copy of this software and associated documentation files (the "Software"),
* to deal in the Software without restriction, including without limitation
* the rights to use, copy, modify, merge, publish, distribute, sublicense,
* and/or sell copies of the Software, and to permit persons to whom
* the Software is furnished to do so, subject to the following conditions:

* The above copyright notice and this permission notice shall be included
* in all copies or substantial portions of the Software.

* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
* EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
* MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
* IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
* DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
* TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
* OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#ifndef _MATERIAL_H_
#define _MATERIAL_H_

#include "Utils/globals.h"
#include "LinearAlgebra/Vector3.h"
#include "LinearAlgebra/Vector2.h"
#include "LightSource/LightSample.h"
#include "Color/Spectrum.h"
#include "Reflectance/Reflectance.h"
#include "RayTracing/Ray.h"

template<int D> class Intersection;
template<int D> class Medium;

/**	Base class for materials. */
template<int D>
class Material
{
protected:
	// Default world media
	static Medium<D> *m_default_medium;
	// Default world index of refraction
	static Real m_default_n;
	// Type of material
	Reflectance::Type m_type;
public:
	Material(Reflectance::Type type = Reflectance::NOTAG):m_type(type){}
	
	/** Computes the local direct reflected (or refracted) light. 
		omega_o is the view (incoming) direction, while omega_i is light (outgoing)*/
	virtual Spectrum f( const VectorN<D> &omega_o, const VectorN<D> &omega_i, const VectorN<D> &normal, const Vector2 &uv ) const = 0;
	/** Virtual function for the PDF */
	virtual Real p(const VectorN<D> &omega_o, const VectorN<D> &omega_i, const VectorN<D> &normal, const Vector2 &uv) const = 0;
	/** Samples a new outgoing direction */
	virtual Spectrum sample_direction( const VectorN<D> &omega_o, VectorN<D> &omega_i, const VectorN<D> &normal, const Vector2 &uv, Real &pdf ) const = 0;
	/** Samples a new outgoing ray in the intersection */	
	virtual Spectrum sample_outgoing_ray( const Intersection<D> &it, Ray<D> &new_ray, Real &pdf ) const = 0;
	
	// TIME RESOLVED MATERIAL FUNCTIONS
	// NOTE: to include re-radiation it's included as parameter the incoming radiance. This is pure crap, but without a reradiation matrix
	// as output (which I have no time now to do) is the only way to include re-radiation. This creates changes in "Path-tracing". But that's, as I said, wrong.
	// In fact, it'd need to adjust path-tracing to be purely recursive, since we cannot trace importance anymore.
	/** Computes the local direct reflected (or refracted) light. 
		omega_o is the view (incoming) direction, while omega_i is light (outgoing)*/
	virtual Spectrum f( const Spectrum &L_i, const VectorN<D> &omega_o, const VectorN<D> &omega_i, const VectorN<D> &normal, const Vector2 &uv, Real &delta_time, Real &pdf ) const 
	{	delta_time = 0.; pdf = 1.f; return L_i*f(omega_o,omega_i,normal,uv); }
	virtual Spectrum f( const VectorN<D> &omega_o, const VectorN<D> &omega_i, const VectorN<D> &normal, const Vector2 &uv, Real &delta_time, Real &pdf ) const 
	{	delta_time = 0.; pdf = 1.f; return f(omega_o,omega_i,normal,uv); }
	/** Virtual function for the PDF */
	virtual Real p(const VectorN<D> &omega_o, const VectorN<D> &omega_i, const VectorN<D> &normal, const Vector2 &uv, Real delta_time)const
	{	return p(omega_o,omega_i,normal,uv); }
	/** Samples a new outgoing direction */
	virtual Spectrum sample_direction( const VectorN<D> &omega_o, VectorN<D> &omega_i, const VectorN<D> &normal, const Vector2 &uv, Real &delta_time, Real &pdf ) const
	{	delta_time = 0.; return sample_direction(omega_o,omega_i,normal,uv,pdf); }
	/** Samples a new outgoing ray in the intersection */	
	virtual Spectrum sample_outgoing_ray( const Intersection<D> &it, Ray<D> &new_ray, Real &delta_time, Real &pdf ) const
	{	delta_time = 0.; return sample_outgoing_ray(it, new_ray, pdf); }


	/** Get the absorption of the material */
	virtual Spectrum get_absorption( const Vector2 &uv ) const = 0;

	virtual Spectrum emit_light( const VectorN<D> &omega_i, const VectorN<D> &normal, const Vector2 &uv )const{return Spectrum(0.);}
	virtual Spectrum emit_light( const VectorN<D> &omega_i, const VectorN<D> &normal, const Vector2 &uv, Real &delta_time, Real &pdf )const{return Spectrum(0.);}
	/** Returns material's type */
	Reflectance::Type get_type() const{ return m_type; }
	/** Returns material's type */
	bool is_type(Reflectance::Type type) const{ return (type & m_type); }
	
	static void set_default_medium(Medium<D> *m){ m_default_medium = m; }
	static void set_default_refraction_index(Real n){ m_default_n = n; }	
}; //Material

typedef Material<3> Material3D;
typedef Material<2> Material2D;

#endif