/* 
* Copyright (C) 2015, Adrian Jarabo (http://giga.cps.unizar.es/~ajarabo/)
*
* Permission is hereby granted, free of charge, to any person obtaining
* a copy of this software and associated documentation files (the "Software"),
* to deal in the Software without restriction, including without limitation
* the rights to use, copy, modify, merge, publish, distribute, sublicense,
* and/or sell copies of the Software, and to permit persons to whom
* the Software is furnished to do so, subject to the following conditions:

* The above copyright notice and this permission notice shall be included
* in all copies or substantial portions of the Software.

* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
* EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
* MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
* IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
* DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
* TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
* OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#ifndef _BSDF_H_
#define _BSDF_H_

#include "Utils/globals.h"
#include "Material/Reflectance/Reflectance.h"
#include "Color/Spectrum.h"
#include "LinearAlgebra/Vector3.h"
#include "LinearAlgebra/Vector2.h"

template<int D> class Ray;
template<int D> class Intersection;

/**	Base virtual class for BSDFs. 
	These model both reflectance (BRDF) and transmitance (BTDF) functions. */
template<int D>
class BSDF
{
protected:
	Reflectance::Type m_type;
public:
	BSDF(Reflectance::Type type = Reflectance::NOTAG):m_type(type){}
	
	/** Virtual function for the BRDF */
	virtual Spectrum f(const VectorN<D> &omega_o, const VectorN<D> &omega_i, const VectorN<D> &normal, const Vector2 &uv) const = 0;
	/** Virtual function for the PDF */
	virtual Real p(const VectorN<D> &omega_o, const VectorN<D> &omega_i, const VectorN<D> &normal, const Vector2 &uv) const = 0;
	/** Virtual function for sampling an outgoing direction, where */
	virtual Spectrum sample_direction(const VectorN<D> &omega_o, VectorN<D> &omega_i, const VectorN<D> &normal, const Vector2 &uv, Real &pdf) const = 0;
	/** Virtual function for sampling an outgoing ray */
	virtual Spectrum sample_outgoing_ray(const Intersection<D> &it, Ray<D> &new_ray, Real &pdf) const = 0;

	// TIME RESOLVED FUNCTIONS (see Material/Material.h)
	/** Virtual function for the BRDF */
	virtual Spectrum f( const Spectrum &L_i, const VectorN<D> &omega_o, const VectorN<D> &omega_i, const VectorN<D> &normal, const Vector2 &uv, Real &delta_time, Real &pdf ) const
	{	delta_time = 0.; pdf = 1.f; return L_i*f(omega_o,omega_i,normal,uv); }
	/** Virtual function for the PDF */
	virtual Real p(const VectorN<D> &omega_o, const VectorN<D> &omega_i, const VectorN<D> &normal, const Vector2 &uv, Real delta_time)const
	{	return p(omega_o,omega_i,normal,uv); }
	/** Virtual function for sampling an outgoing direction, where */
	virtual Spectrum sample_direction( const VectorN<D> &omega_o, VectorN<D> &omega_i, const VectorN<D> &normal, const Vector2 &uv, Real &delta_time, Real &pdf ) const
	{	delta_time = 0.; return sample_direction(omega_o,omega_i,normal,uv,pdf); }
	/** Virtual function for sampling an outgoing ray */
	virtual Spectrum sample_outgoing_ray( const Intersection<D> &it, Ray<D> &new_ray, Real &delta_time, Real &pdf ) const
	{	delta_time = 0.; return sample_outgoing_ray(it, new_ray, pdf); }

	/** Get the absorption of the BSDF */
	virtual Spectrum get_absorption( const Vector2 &uv ) const = 0;
	/** Returns material's type */
	Reflectance::Type get_type() const{ return m_type; }
	/** Returns material's type */
	bool is_type(Reflectance::Type type) const{ return (type & m_type); }
}; //BSDF
 
typedef BSDF<3> BSDF3D;
typedef BSDF<2> BSDF2D;

#endif //_BSDF_H_