/* 
* Copyright (C) 2015, Adrian Jarabo (http://giga.cps.unizar.es/~ajarabo/)
*
* Permission is hereby granted, free of charge, to any person obtaining
* a copy of this software and associated documentation files (the "Software"),
* to deal in the Software without restriction, including without limitation
* the rights to use, copy, modify, merge, publish, distribute, sublicense,
* and/or sell copies of the Software, and to permit persons to whom
* the Software is furnished to do so, subject to the following conditions:

* The above copyright notice and this permission notice shall be included
* in all copies or substantial portions of the Software.

* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
* EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
* MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
* IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
* DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
* TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
* OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#ifndef _LAMBERTIAN_H_
#define _LAMBERTIAN_H_

#include "Material/Reflectance/BSDF.h"
#include "RayTracing/Ray.h"
#include "RayTracing/Intersection.h"
#include <math.h>

/**	Simple Lambertian BRDF. */
template<int D>
class Lambertian: public BSDF<D>
{
	//Specular coefficients
	Spectrum Kd;
public:
	Lambertian(Spectrum _Kd = Spectrum(.67)):Kd(_Kd),BSDF<D>(Reflectance::DIFFUSE){}
	
	Spectrum f(const VectorN<D> &omega_o, const VectorN<D> &omega_i, const VectorN<D> &normal, const Vector2 &uv)const;
	Real p(const VectorN<D> &omega_o, const VectorN<D> &omega_i, const VectorN<D> &normal, const Vector2 &uv)const;
	
	Spectrum sample_direction(const VectorN<D> &omega_o, VectorN<D> &omega_i, const VectorN<D> &normal, const Vector2 &uv, Real &pdf) const;
	Spectrum sample_outgoing_ray(const Intersection<D> &it, Ray<D> &new_ray, Real &pdf) const;

	Spectrum get_absorption( const Vector2 &uv ) const
	{	return Spectrum(1.) - Kd;	}
	
}; //Lambertian

//------------------------------------------
//Specializations for 3D
template<>
Spectrum Lambertian<3>::f(const Vector3 &omega_o, const Vector3 &omega_i, const Vector3 &normal, const Vector2 &uv)const
{
	return Kd*(1.f/M_PI);
}

template<>
Real Lambertian<3>::p(const Vector3 &omega_o, const Vector3 &omega_i, const Vector3 &normal, const Vector2 &uv)const
{
	Real cos_theta = dot(omega_i, normal);
	return (cos_theta/static_cast<Real>(M_PI));
}

template<>
Spectrum Lambertian<3>::sample_direction(const Vector3 &omega_o, Vector3 &omega_i, const Vector3 &normal, const Vector2 &uv, Real &pdf) const
{
	//Sample hemisphere cosine weighted-> Need to be redone...
	Real epsilon1 = static_cast<Real>(rand())/static_cast<Real>(RAND_MAX);
	while( epsilon1 <= 0. )
		epsilon1 = static_cast<Real>(rand())/static_cast<Real>(RAND_MAX);

	Real epsilon2 = static_cast<Real>(rand())/static_cast<Real>(RAND_MAX);
	while( epsilon2 <= 0.0 || epsilon2 >= 1.)
		epsilon2 = static_cast<Real>(rand())/static_cast<Real>(RAND_MAX);

	Real phi = 2*M_PI*epsilon1;
	Real cos_theta = sqrt(1-epsilon2*epsilon2); // CHECK COSINE WEIGHTED SAMPLING!!
	Real sin_theta = epsilon2;

	pdf = (cos_theta/static_cast<Real>(M_PI));
	Vector3 dir(cos(phi)*sin_theta, cos_theta, sin(phi)*sin_theta);

	omega_i = dir.transform_matrix_to(Vector3(0,1,0), normal);

	return f(-omega_o, -omega_i, normal, uv);
}

template<>
Spectrum Lambertian<3>::sample_outgoing_ray(const Intersection3D &it, Ray3D &new_ray, Real &pdf) const
{
	Vector3 new_omega_i;
	Spectrum refl = sample_direction(it.get_ray().get_direction(), new_omega_i, it.get_normal(),it.get_uv(), pdf);

	new_ray = Ray3D(it.get_position(), new_omega_i, true, it.get_ray().get_level()+1, 
		it.get_ray().get_refraction_index(), it.get_ray().get_medium());
	
	new_ray.shift();

	return refl;
}

//------------------------------------------
//Specializations for 2D
template<>
Spectrum Lambertian<2>::f(const Vector2 &omega_o, const Vector2 &omega_i, const Vector2 &normal, const Vector2 &uv)const
{
	//return (dot(-omega_o, normal) > 0.)?(Kd*0.5f):0.;
	return (Kd*0.5f);
}
template<>
Real Lambertian<2>::p(const Vector2 &omega_o, const Vector2 &omega_i, const Vector2 &normal, const Vector2 &uv)const
{
	Real cos_theta = dot(omega_i, normal);
	return (cos_theta*0.5);
}
template<>
Spectrum Lambertian<2>::sample_direction(const Vector2 &omega_o, Vector2 &omega_i, const Vector2 &normal, const Vector2 &uv, Real &pdf) const
{
	//Sample hemisphere cosine weighted-> Need to be redone...
	
	Real epsilon2 = static_cast<Real>(rand())/static_cast<Real>(RAND_MAX);
	while( epsilon2 < 0. || epsilon2 == 1.)
		epsilon2 = static_cast<Real>(rand())/static_cast<Real>(RAND_MAX);
	
	Real cos_theta = sqrt(1. - epsilon2*epsilon2); //sample right or left of the semicircle
	Real sin_theta = ((rand() % 2)?1.:-1.)*epsilon2;
	
	pdf = 0.5*cos_theta;
	
	Vector2 dir(sin_theta, cos_theta);
	
	omega_i = dir.transform_matrix_to(Vector2(0,1), normal);
	
	return f(-omega_o, -omega_i, normal, uv);
}

template<>
Spectrum Lambertian<2>::sample_outgoing_ray(const Intersection2D &it, Ray2D &new_ray, Real &pdf) const
{
	Vector2 new_omega_i;
	Spectrum refl = sample_direction(it.get_ray().get_direction(), new_omega_i, it.get_normal(),it.get_uv(), pdf);
	
	new_ray = Ray2D(it.get_position(), new_omega_i, true, it.get_ray().get_level()+1, 
				  it.get_ray().get_refraction_index(), it.get_ray().get_medium());
	
	new_ray.shift();
	
	return refl;
}



typedef Lambertian<3> Lambertian3D;
typedef Lambertian<2> Lambertian2D;

#endif //_LAMBERTIAN_H_