/* 
* Copyright (C) 2015, Adrian Jarabo (http://giga.cps.unizar.es/~ajarabo/)
*
* Permission is hereby granted, free of charge, to any person obtaining
* a copy of this software and associated documentation files (the "Software"),
* to deal in the Software without restriction, including without limitation
* the rights to use, copy, modify, merge, publish, distribute, sublicense,
* and/or sell copies of the Software, and to permit persons to whom
* the Software is furnished to do so, subject to the following conditions:

* The above copyright notice and this permission notice shall be included
* in all copies or substantial portions of the Software.

* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
* EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
* MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
* IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
* DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
* TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
* OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#ifndef _PHONG_H_
#define _PHONG_H_

#include "Material/Reflectance/BSDF.h"
#include "RayTracing/Ray.h"
#include "RayTracing/Intersection.h"
#include <math.h>

/**	Physically-based Blinn-Phong BRDF:
	Blinn, J. 1977. "Models of light reflection for computer synthesized 
	pictures". In SIGGRAPH 1977.
	http://dl.acm.org/citation.cfm?doid=563858.563893	
	[TODO] Develop 2D versions of the BRDF, right now only 3D. */
class Phong: public BSDF<3>
{
	//Specular and Diffuse coefficients
	Spectrum Ks, Kd;
	//Specular sharpness (glossiness)
	Real Ns;
public:
	Phong():Kd(.67),Ks(1.),Ns(0.),BSDF<3>(Reflectance::GLOSSY){}
	Phong(const Spectrum &_Kd, const Real _Ns): Kd(_Kd),Ks(1.),Ns(_Ns),BSDF<3>(Reflectance::GLOSSY){}
	Phong(const Spectrum &_Kd, const Spectrum &_Ks, const Real _Ns): Kd(_Kd), Ks(_Ks), Ns(_Ns),BSDF<3>(Reflectance::GLOSSY){}
	
	Spectrum f(const Vector3 &omega_o, const Vector3 &omega_i, const Vector3 &normal, const Vector2 &uv)const
	{
		Vector3 half_vector = normalize(-omega_i+omega_o);
		Real clamped_dot = dot(normal, half_vector); clamped_dot = (clamped_dot>0.f)?clamped_dot:0.f;
		Real cosine_lobe = pow(clamped_dot, Ns);

		return (Kd + Ks*(.5f*(Ns+2.f)*cosine_lobe))*(1.f/M_PI);
	}
	Real p(const Vector3 &omega_o, const Vector3 &omega_i, const Vector3 &normal, const Vector2 &uv) const
	{
		//Sample hemisphere cosine weighted-> Need to be redone...
		return (dot(omega_i,normal)/static_cast<Real>(M_PI));
	}
	Spectrum sample_direction(const Vector3 &omega_o, Vector3 &omega_i, const Vector3 &normal, const Vector2 &uv, Real &pdf) const
	{
		//Sample hemisphere cosine weighted-> Need to be redone...
		Real epsilon1 = static_cast<Real>(rand())/static_cast<Real>(RAND_MAX);
		Real epsilon2 = static_cast<Real>(rand())/static_cast<Real>(RAND_MAX);
		while( epsilon2 == 0. )
			epsilon2 = static_cast<Real>(rand())/static_cast<Real>(RAND_MAX);

		Real phi = 2*M_PI*epsilon1;
		Real cos_theta = sqrt(epsilon2);
		Real sin_theta = sqrt(1.-cos_theta*cos_theta);

		pdf = (cos_theta/static_cast<Real>(M_PI));
		
		Vector3 dir(cos(phi)*sin_theta, cos_theta, sin(phi)*sin_theta);

		omega_i = dir.transform_matrix_to(Vector3(0,1,0), normal);

		return f(-omega_o, -omega_i, normal, uv);
	}
	Spectrum sample_outgoing_ray(const Intersection3D &it, Ray3D &new_ray, Real &pdf) const
	{
		Vector3 new_omega_i;
		Spectrum refl = sample_direction(it.get_ray().get_direction(), new_omega_i, it.get_normal(),it.get_uv(), pdf);

		new_ray = Ray3D(it.get_position(), new_omega_i, true, it.get_ray().get_level()+1, 
			it.get_ray().get_refraction_index(), it.get_ray().get_medium());
		
		new_ray.shift();
		
		return refl;
	}

	Spectrum get_absorption( const Vector2 &uv ) const
	{	return Kd;	}
	
}; //Phong

#endif //_PHONG_H_