/* 
* Copyright (C) 2015, Adrian Jarabo (http://giga.cps.unizar.es/~ajarabo/)
*
* Permission is hereby granted, free of charge, to any person obtaining
* a copy of this software and associated documentation files (the "Software"),
* to deal in the Software without restriction, including without limitation
* the rights to use, copy, modify, merge, publish, distribute, sublicense,
* and/or sell copies of the Software, and to permit persons to whom
* the Software is furnished to do so, subject to the following conditions:

* The above copyright notice and this permission notice shall be included
* in all copies or substantial portions of the Software.

* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
* EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
* MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
* IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
* DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
* TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
* OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#ifndef _PHASE_FUNCTION_H_
#define _PHASE_FUNCTION_H_

#include "Sampling/Sampling.h"
#include "Color/Spectrum.h"
#include "LinearAlgebra/Vector3.h"
#include "LinearAlgebra/Vector2.h"

namespace PhaseFunction
{
	//-------------------------------------------------------------------------------------------------
	// Base virtual phase function. 
    template <int D>
	class PhaseFunction
	{
		Real m_mean_time; //this should be changed for some more intuitive parameter
		Real m_inv_mean_time;
	protected:
		// Sample the time delay
		virtual Real sample_time_delay(Real &pdf) const
		{	
			Real epsilon1, dist, pf_pdf;
            if (m_mean_time == 0)
            {
                pdf = 1.;
                return 0.;
            }
            else
			{
                epsilon1 = RNG::StdRandom::get_real();
                Real delta_time = -logf(epsilon1)/m_inv_mean_time;
                pdf = exp(-m_inv_mean_time*delta_time)*m_inv_mean_time;
				return delta_time;
            }
		}
		virtual Real pdf_time_delay(const Real delta_time)const
		{
			if (m_mean_time == 0)
                return 1.;

			return exp(-m_inv_mean_time*delta_time)*m_inv_mean_time;
		}
	public:
		PhaseFunction(const Real mean_time): m_mean_time(mean_time), m_inv_mean_time(mean_time>0?1./mean_time:0.){}

		virtual Spectrum sample_direction(const VectorN<D>& wi, VectorN<D> &wo, Real &pdf) const = 0;
		virtual Spectrum operator()(const VectorN<D>& wi, const VectorN<D>& wo) const = 0;
        virtual Real p(const VectorN<D>& wi, const VectorN<D> &wo) const = 0;

        //TIME RESOLVED PHASE FUNCTIONS
        virtual Spectrum sample_direction(const VectorN<D>& wi, VectorN<D> &wo, 
                                          Real &delta_time, Real &pdf) const
		{        
			Spectrum r = sample_direction(wi,wo,pdf);
			
			Real pdf_time;
			delta_time = sample_time_delay(pdf_time);
			pdf *= pdf_time;

			return r*pdf_time;
		}

        virtual Spectrum operator()(const VectorN<D>& wi, const VectorN<D>& wo, 
                                    Real &delta_time, Real &pdf) const
        {
			delta_time = sample_time_delay(pdf);

			return pdf*(*this)(wi,wo);
		};

		virtual Real p(const VectorN<D>& wi, const VectorN<D> &wo, const Real delta_time)const
		{	
			return pdf_time_delay(delta_time)*p(wi,wo); 
		}
	}; // class PhaseFunction

}; //namespace PhaseFunction

#endif //_PHASE_FUNCTION_H_