/* 
* Copyright (C) 2015, Adrian Jarabo (http://giga.cps.unizar.es/~ajarabo/)
*
* Permission is hereby granted, free of charge, to any person obtaining
* a copy of this software and associated documentation files (the "Software"),
* to deal in the Software without restriction, including without limitation
* the rights to use, copy, modify, merge, publish, distribute, sublicense,
* and/or sell copies of the Software, and to permit persons to whom
* the Software is furnished to do so, subject to the following conditions:

* The above copyright notice and this permission notice shall be included
* in all copies or substantial portions of the Software.

* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
* EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
* MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
* IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
* DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
* TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
* OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#ifndef _RAYLEIGH_PHASE_FUNCTION_H_
#define _RAYLEIGH_PHASE_FUNCTION_H_

#include "Media/PhaseFunction/PhaseFunction.h"
#include "RayleighScattering.h"

namespace PhaseFunction
{
	//-------------------------------------------------------------------------------------------------
	// Rayleigh phase function. 
	// Computes the rayleigh scattering phase function, based on the accurate formula by Chandrasekhar
	// including molecular anisotropy and the depolarization term [1,2].
	//-------------------------------------------------------------------------------------------------
	// Known issues:
	// 1-	The sampling method used is just an spherical sampling, which is known to be a poor sampling
	//		strategy for Rayleigh phase function [3].
	//
	// [1]	S. Chandrasekhar 1960. Radiative Transfer. Dover.
	// [2]	A. Bucholtz 1995. Rayleigh-scattering calculations for the terrestrial atmosphere. Applied
	//		Optics, Vol. 34(15).
	//		http://augerlal.lal.in2p3.fr/pmwiki/uploads/Bucholtz.pdf
	// [3]	J. Frisvad 2011. Importance sampling the Rayleigh phase function. In JOSA: Optics, Image 
	//		Science, and Vision, 28(12)
	//		http://orbit.dtu.dk/fedora/objects/orbit:40641/datastreams/file_6314521/content
	template<int D>
	class Rayleigh: public PhaseFunction<D>
	{
	public:
		Rayleigh(): 
	        PhaseFunction<D>(0.) 
        {}
				
		Spectrum sample_direction(const VectorN<D> &wi, VectorN<D> &wo, Real &pdf) const;
		Spectrum operator()(const VectorN<D>& wi, const VectorN<D>& wo) const;
		Real p(const VectorN<D>& wi, const VectorN<D> &wo) const;


		//TIME RESOLVED PHASE FUNCTIONS
        Spectrum sample_direction(const VectorN<D>& wi, VectorN<D> &wo, 
                                          Real &delta_time, Real &pdf) const
		{  
			delta_time = 0.; pdf = 1;
			return sample_direction(wi,wo,pdf);
		}

        virtual Spectrum operator()(const VectorN<D>& wi, const VectorN<D>& wo, 
                                    Real &delta_time, Real &pdf) const
        {
			delta_time = 0.; pdf = 1;
			return operator()(wi,wo);
		}

		virtual Real p(const VectorN<D>& wi, const VectorN<D> &wo, const Real delta_time)const
		{	
			return p(wi,wo);
		}
	}; // class Rayleigh

	template<>
	Spectrum Rayleigh<2>::sample_direction(const VectorN<2> &wi, VectorN<2> &wo, Real &pdf) const
	{
		throw("Invalid dimensions: Rayleigh scattering only accepts N=3\n");
	}

	template<>
	Spectrum Rayleigh<3>::sample_direction(const Vector3& wi, Vector3 &wo, Real &pdf) const
	{
		Sampling::spherical(wo,pdf);
		return (*this)(wi,wo);
	}

	template<>
	Spectrum Rayleigh<2>::operator()(const Vector2& wi, const Vector2& wo) const
	{
		throw("Invalid dimensions: Rayleigh scattering only accepts N=3\n");
	}


	template<>
	Spectrum Rayleigh<3>::operator()(const Vector3& wi, const Vector3& wo) const
	{
		Real cos_theta = dot(wi,wo);
		Spectrum gamma = RayleighScattering::gamma();
		return 3./(4.+8.*gamma)*(1.+3.*gamma + (1-gamma)*cos_theta*cos_theta);
	}
    
	template<>
	Real Rayleigh<2>::p(const VectorN<2>& wi, const VectorN<2> &wo) const
	{
		throw("Invalid dimensions: Rayleigh scattering only accepts N=3\n");
	}

	template<>
	Real Rayleigh<3>::p(const VectorN<3>& wi, const VectorN<3> &wo) const
	{
		return (.25*M_1_PI);
	}

	

}; //namespace PhaseFunction



#endif //_RAYLEIGH_PHASE_FUNCTION_H_