/* 
* Copyright (C) 2015, Adrian Jarabo (http://giga.cps.unizar.es/~ajarabo/)
*
* Permission is hereby granted, free of charge, to any person obtaining
* a copy of this software and associated documentation files (the "Software"),
* to deal in the Software without restriction, including without limitation
* the rights to use, copy, modify, merge, publish, distribute, sublicense,
* and/or sell copies of the Software, and to permit persons to whom
* the Software is furnished to do so, subject to the following conditions:

* The above copyright notice and this permission notice shall be included
* in all copies or substantial portions of the Software.

* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
* EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
* MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
* IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
* DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
* TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
* OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#ifndef _RENDERENGINE_H_
#define _RENDERENGINE_H_

//#include <time.h>

#include "Utils/globals.h"

#include "Utils/Timer.h"

#include "LinearAlgebra/Vector3.h"
#include "Sampling/Sampler.h"
#include "Film/Film.h"
#include "RayTracing/World.h"
#include "Camera/Camera.h"
#include "Integrator/Integrator.h"
#include "Integrator/RadianceSample.h"

#include <math.h>
#include <iostream>


/** Base class implementing the simplest render engine. 
	It contains the rendering main loop, and supports 
	non-progressive, steady-state rendering.		*/
template<int D>
class RenderEngine
{
public:
	/** Constructor. */
 	RenderEngine()
	{	/*srand((unsigned)time(0));*/	}
	
	/** Function that implements the rendering main loop.
		It gets as input the name of the image to be rendered,
		and all the different elements needed to render the 
		scene, including the scene itself, the integrator, and 
		the camera, film and sampler. */
	virtual void render( const char *name, World<D> *world, Integrator<D> *integrator,
						Camera<D> *camera, Film *film, Sampler *sampler )const;
}; //RenderEngine

//JULIO: Added volume integrator to the call and to the implementation
template<int D>
void RenderEngine<D>::render( const char *name, World<D> *world, Integrator<D> *integrator,
							 Camera<D> *camera, Film *film, Sampler *sampler )const
{
	// ----------------------------------------------------------------------
	// If no integrator has been defined, set error...	
	if( !integrator )
	{		throw("Error: Need to define an integrator for rendering\n"); }

	// ----------------------------------------------------------------------
	// If needed, prepare the scene to be render ...
	cout << "Preparing scene to be rendered...\r";
	Timer timer;timer.start();
	
	// Construct acceleration structures and such
	world->freeze();
	{
		Real secs = timer.get_secs();
		int hours = static_cast<int>(secs)/3600; secs -= hours*3600;
		int minutes = static_cast<int>(secs)/60; secs -= minutes*60;
		std::cout << "Prepared scene to render: \t["<<hours<<":"<<minutes<<":"<<secs<<"]             \n"; 
	}

	// ----------------------------------------------------------------------
	// If needed, preprocess integrator ...
	cout << "Preprocessing integrator...\r";
	timer.start();

	integrator->preprocess();
	{
		Real secs = timer.get_secs();
		int hours = static_cast<int>(secs)/3600; secs -= hours*3600;
		int minutes = static_cast<int>(secs)/60; secs -= minutes*60;
		std::cout << "Preprocessed integrator: \t["<<hours<<":"<<minutes<<":"<<secs<<"]             \n"; 
	}
	

	// ----------------------------------------------------------------------
	// Go rendering!
	cout << "Rendering ...\r"; 
	timer.start();

	//Main Loop: Loop over all samples in the image
	Sample film_sample;
	int nb_samples = sampler->get_nb_samples(), curr_sample = 0;
	FILE *fout = fopen("log.txt", "w+");
	
	while( sampler->get_next_sample(film_sample) )
	{
		// Update timer
		if(!(curr_sample%2000))
		{
			Real secs = timer.get_secs();
			int hours = static_cast<int>(secs)/3600; secs -= hours*3600;
			int minutes = static_cast<int>(secs)/60; secs -= minutes*60;
			std::cout << "Rendering ..."
				<<static_cast<Real>(curr_sample)/nb_samples*100<<"%: \t["<<hours<<":"<<minutes<<":"<<secs<<"]            "<<"\r"<<std::flush;
		}
		curr_sample++;
				
		// Get the camera sample ray...		
		VectorN<D-1> ic;
		//= film->window_coords2image_coords(film_sample.position);
		Vector2 meh = film->window_coords2image_coords(film_sample.position);
		
		if (D == 2)
		{
			ic[0] = meh[0];
		}
		else 
		{
			ic[0] = meh[0]; ic[1] = meh[1];
			
			//ic = film->window_coords2image_coords(film_sample.position);
		}
		
		Real pdf_camera = 0.;
		Ray<D> r;
		r = camera->get_ray(ic, pdf_camera);

		if( pdf_camera == 0. )
			continue;

		//printf("Ray: Dir[%f,%f,%f]\n", r.get_direction()[0],r.get_direction()[1],r.get_direction()[2]);
		film_sample.weight /= pdf_camera;
		
		// ... trace the ray and compute incoming radiance...
		RadianceSample radiance_sample = world->Li(r, integrator);
		// ... and store it in the film
		film->add_sample(film_sample, radiance_sample);
	}

	//----------------------------------------------------------------------
	// End rendering, stop timer
	{
		Real secs = timer.get_secs();
		int hours = static_cast<int>(secs)/3600; secs -= hours*3600;
		int minutes = static_cast<int>(secs)/60; secs -= minutes*60;
		cout << "Rendering ..."
					<<"[DONE]: \t["<<hours<<":"<<minutes<<":"<<secs<<"]                        "<<"\n"; 
	}
	
	// Save the image
	film->write(name);
}


#endif //_RENDERENGINE_H_

