/* 
* Copyright (C) 2015, Adrian Jarabo (http://giga.cps.unizar.es/~ajarabo/)
*
* Permission is hereby granted, free of charge, to any person obtaining
* a copy of this software and associated documentation files (the "Software"),
* to deal in the Software without restriction, including without limitation
* the rights to use, copy, modify, merge, publish, distribute, sublicense,
* and/or sell copies of the Software, and to permit persons to whom
* the Software is furnished to do so, subject to the following conditions:

* The above copyright notice and this permission notice shall be included
* in all copies or substantial portions of the Software.

* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
* EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
* MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
* IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
* DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
* TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
* OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#ifndef _TRANSIENT_RENDERER_H_
#define _TRANSIENT_RENDERER_H_

#include "RenderEngine.h"
#include "LightSource/LightSample.h"
#include "Film/StreakCameraFilm.h"
#include "Utils/Timer.h"

#include <math.h>
#include <iostream>
#include <list>

/** Class implementing transient rendering engine. 
	It contains the rendering main loop, and supports 
	non-progressive time-resolved rendering. 
	Note that it needs StreakCameraFilm to work properly. */
template<int D>
class TransientRenderer: public RenderEngine<D>
{
	FILE *f_log;
	bool time_sampling;
public:
	TransientRenderer(FILE *_f_log, bool _time_sampling = false):f_log(_f_log), time_sampling(_time_sampling)
	{}	
	TransientRenderer(){}	
	bool sensor_mode;
	virtual void render( const char *name, World<D> *world, Integrator<D> *integrator, 
						Camera<D> *camera, Film *film, Sampler *sampler)const;
	void set_sensor_mode (bool _sensor_mode) {sensor_mode = _sensor_mode; }
}; //TransientRenderer

template<int D>
void TransientRenderer<D>::render(const char *name, World<D> *world, Integrator<D> *integrator,
								Camera<D> *camera, Film *film, Sampler *sampler )const
{	
	if (f_log) 
	{
		fprintf(f_log, "RENDER\n");
		fprintf(f_log, "======\n");
	}
	// ----------------------------------------------------------------------
	// If no integrator has been defined, use the default...	
	if( !integrator )
	{		throw("Error: Need to define an integrator for rendering\n"); }
	
	// This rendered can only render into streak films. Thus, it is needed to
	// test it.
	if( static_cast<StreakCameraFilm*>(film) == 0 )
		throw("Transient Renderer needs an StreakCameraFilm");


	// ----------------------------------------------------------------------
	// If needed, prepare the scene to be render ...
	cout << "Preparing scene to be rendered...\r";
	Timer timer;timer.start();
	// Construct acceleration structures and such
	world->freeze();
	{
		Real secs = timer.get_secs();
		int hours = static_cast<int>(secs)/3600; secs -= hours*3600;
		int minutes = static_cast<int>(secs)/60; secs -= minutes*60;
		std::cout << "Prepared scene to render: \t["<<hours<<":"<<minutes<<":"<<secs<<"]             \n"; 
		if (f_log) 
			fprintf(f_log, "Prepared scene to render: %d:%d:%f\n", hours, minutes, secs);
	}

	// ----------------------------------------------------------------------
	// If needed, preprocess integrator ...
	cout << "Preprocessing integrator...\r";
	timer.start();

	integrator->preprocess();
	{
		Real secs = timer.get_secs();
		int hours = static_cast<int>(secs)/3600; secs -= hours*3600;
		int minutes = static_cast<int>(secs)/60; secs -= minutes*60;
		std::cout << "Preprocessed integrator: \t["<<hours<<":"<<minutes<<":"<<secs<<"]             \n"; 
		if (f_log) 
			fprintf(f_log, "Preprocessed integrator: %d:%d:%f\n", hours, minutes, secs);

	}

	// ----------------------------------------------------------------------
	// Go rendering!
	cout << "Rendering ...\r"; 
	timer.start();

	//Loop over all samples in the image
	Sample film_sample;
	int nb_samples = sampler->get_nb_samples(), curr_sample = 0;

	while( sampler->get_next_sample(film_sample) )
	{
		// Update timer
		if(!(curr_sample%30))
		{
			Real secs = timer.get_secs();
			int hours = static_cast<int>(secs)/3600; secs -= hours*3600;
			int minutes = static_cast<int>(secs)/60; secs -= minutes*60;
			std::cout << "Rendering ..."
				<<static_cast<Real>(curr_sample)/nb_samples*100<<"%: \t["<<hours<<":"<<minutes<<":"<<secs<<"]             "<<"\r"<<std::flush; 
		}
		curr_sample++;

		VectorN<D-1> ic;
		
		Vector2 meh = film->window_coords2image_coords(film_sample.position);
		
		if (D == 2) ic[0] = meh[0]; 
		else { ic[0] = meh[0]; ic[1] = meh[1];}
		
		// Get the camera sample ray...
		Ray<D> r = camera->get_ray(ic);
		// ... trace the ray, compute time-resolved samples, and store it in a list...
		std::list<RadianceSample> samples;
		
		if (time_sampling)
            world->Li(r, integrator, samples, film->get_time_length(),film->get_time_resolution());
		else
            world->Li(r, integrator, samples, film->get_exposure_time(),0);

		// ... and finally store the samples in the film
		film->add_samples(film_sample, samples);
	}

	//----------------------------------------------------------------------
	// End rendering, stop timer
	{
		Real secs = timer.get_secs();
		int hours = static_cast<int>(secs)/3600; secs -= hours*3600;
		int minutes = static_cast<int>(secs)/60; secs -= minutes*60;
		cout << "Rendering ..."
					<<"[DONE]: \t["<<hours<<":"<<minutes<<":"<<secs<<"]                        "<<"\n"; 
		if (f_log) 
		{
			fprintf(f_log, "Total time: %d:%d:%f\n", hours, minutes, secs);
			fflush(f_log);
		}
	}
	
	// Save the image
	film->write(name);
}

#endif //_TRANSIENT_RENDERER_H_