#include <iostream>
#include "Utils/globals.h"

#include "Camera/Camera.h"
#include "Camera/PinholeCamera.h"

#include "RayTracing/World.h"

#include "Renderer/TransientRenderer.h"
#include "Film/MultibounceStreakCameraFilm.h"

#include "Geometry/Primitives/3D/Sphere.h"
#include "Geometry/Primitives/3D/Plane.h"
#include "Geometry/Aggregate/Mesh.h"
#include "Geometry/Aggregate/BVH.h"
#include "Geometry/ObjectPointer.h"

#include "Material/UberMaterial.h"
#include "Material/Reflectance/Phong.h"
#include "Material/Reflectance/Lambertian.h"
#include "Material/Reflectance/Transmissive.h"
#include "Material/Reflectance/Specular.h"
#include "Material/Reflectance/FresnelDielectric.h"

#include "Material/LambertianTextured.h"

#include "LightSource/PointLightSource.h"
#include "LightSource/DirectionalLightSource.h"
#include "LightSource/CosineLightSource.h"
#include "LightSource/SpotLightSource.h"
#include "LightSource/RectangularAreaLightSource.h"

#include "Filter/GaussianFilter.h"
#include "Filter/UberFilter.h"

#include "Sampling/StratifiedSampler.h"
#include "Utils/RandomNumbers.h"

#include "Integrator/BidirectionalPathTracing.h"

#include "Media/HomogeneousMedium.h"
#include "Media/PhaseFunction/IsotropicPhaseFunction.h"
#include "Media/PhaseFunction/HenyeyGreensteinPhaseFunction.h"



//#include "Beam.h"
#include "Utils/Timer.h"

#define DIM 3

using namespace std;

// Material loader:
Material3D *load_material( int &i, int argc, char* argv[], Material3D *default_material )
{
	Material3D *mesh_mat= default_material;
	
	if( !strcmp("-lambertian",argv[i+1]))
	{	i+=2; UberMaterial3D *um = new UberMaterial3D(); um->add_bsdf(new Lambertian3D(Spectrum(atof(argv[i]),atof(argv[i+1]),atof(argv[i+2])))); mesh_mat = um;i+=2;}
	else if( !strcmp("-phong",argv[i+1]))
	{	i+=2; UberMaterial3D *um = new UberMaterial3D(); um->add_bsdf(new Phong(Spectrum(atof(argv[i])),Spectrum(atof(argv[i+1])),atof(argv[i+2]))); mesh_mat = um; i+=2;}
	else if( !strcmp("-dielectric",argv[i+1]))
	{	i+=2; UberMaterial3D *um = new UberMaterial3D(); um->add_bsdf(new FresnelDielectric3D(Spectrum(atof(argv[i])), atof(argv[i+1]))); mesh_mat = um; i+=1; }
	else if( !strcmp("-lambertian-textured",argv[i+1]))
	{	i+=2; LambertianTextured3D *lt = new LambertianTextured3D(Spectrum(atof(argv[i])),argv[i+1]); mesh_mat = lt; i+=1; }

	return mesh_mat;
}

int main(int argc, char* argv[])
{	
	Timer timer;

    
	// Prints all input arguments.
	for (int a=0; a < argc; a++)
        fprintf(stdout, "argv[%d]: %s\n", a,argv[a]);
    fprintf(stdout, "\n");
    
	// CAMERA and SENSOR PARAMETERS
	Vector3 camera_position(5,5,5) ;
	Vector3 looking_at(5,0,0);
	Vector3 up(0,1,0);
	// Ortho camera:
	Vector2 o_camera_pos (0, 0), o_camera_dir(1, 0);
	Real o_camera_w = 1;
	
	Real view_plane_dist = 1;
	Real focal_distance = 2.6;
	
	Vector2 sensor_center, sensor_dims; 
	
	//LIGHT PARAMETERS
	VectorN<DIM> light_pos(0.0f);
	Spectrum light_ints(1.0f);
	// SPOT LIGHT PARAMETERS
	VectorN<DIM> spot_light_target;
	Real spot_light_angle;

	//FILTER
	Real t_sigma_x=5, t_sigma_y=.9;
	
	Spectrum grey(0.3f);
	Spectrum grey_50perc(0.5f);
	
	//FILM
	int width = 256, height = (DIM==2)?1:width, time = 512;
	int scanline = -1;
	Real time_resolution = 100., film_offset = 0.0;
	int sqrt_spp = 2;
	bool export_depth = false, export_variance = false;
	bool single_pixel = false; 
	int single_pixel_x = -1, single_pixel_y = -1;


	int samples_hemisphere = 1024;
	
	Vector2 image_coordinates_vpl(27,128-76);
	// WORLD
	World<DIM>*  w = new World<DIM>();

	//FILES AND FOLDERS
	char *name_file = NULL, *default_name_file = "name_file", *name_folder = "./";
	char *svg_file = NULL;
	Real svg_scale = 1.;
	name_file = default_name_file;
	
	char *log_file = NULL, *default_log_file = "name_file.txt";
 
	
	std::vector< Object3DPointer > objects;
	std::vector< Triangle > triangles;
	
	bool transient = true, is_multibounce_streak = false, camera_unwarp = false, 
		 time_sampling = true, sensor_mode = false;
	unsigned int vol_path_tracing_samples = 1024, new_seed = 0, nb_multibounces = 3;

	//MEDIUM PARAMETERS
	Spectrum medium_sigma_a(0.), medium_sigma_s(0.); 
	Real hg_g = 0.;
	Scattering::Level scat_level = Scattering::ALL;
	Real pf_time = 0.; //time_resolution;

	// First look for material
	for( int i=1; i<argc; ++i)
	{
		if( !strcmp("-lambertian-rho", argv[i])) {++i; grey = Spectrum(atof(argv[i])); }
	}
	Lambertian3D rho_grey(grey);
	UberMaterial3D m_grey; m_grey.add_bsdf(&rho_grey);
	
	int max_nb_bounces = 20;


	for( int i=1; i<argc; ++i)
	{	
		// Mesh
		if( !strcmp("-name-mesh", argv[i]) ) 
		{	
			++i; char *mesh_name = argv[i];

			Material3D *mat = load_material(i,argc,argv, static_cast<Material3D*>(&m_grey));

			Mesh *m = new Mesh(mesh_name,mat);
			//objects.push_back(Object3DPointer(m));
			m->push_primitives(triangles);

			if( triangles.size() > 10 )
			{
				BVH<Triangle,3> *bvh = new BVH<Triangle,3>(triangles);
				objects.push_back(Object3DPointer(bvh));
				//w->add_object(bvh);
			}
			else
				objects.push_back(Object3DPointer(m));
			triangles.clear();
		}
		// Sphere
		if( !strcmp("-sphere", argv[i]) )
		{
			++i; Vector3 p(atof(argv[i]),atof(argv[i+1]),atof(argv[i+2])); i+=3; Real r = atof(argv[i]);
			
			Material3D *mat = load_material(i,argc,argv, static_cast<Material3D*>(&m_grey));
			objects.push_back(Object3DPointer(new Sphere(p,r,mat)));
		}
		
		if( !strcmp("-folder-name", argv[i]) ) {++i; name_folder = argv[i];}
		if( !strcmp("-film-name", argv[i]) ) {++i; name_file = argv[i];}
		if( !strcmp("-film-size-x",argv[i]) ) {++i; width = atoi(argv[i]); height = (DIM == 2)?1:width;}
		if( !strcmp("-film-size-y",argv[i]) ) {++i; height = atoi(argv[i]);}
		if( !strcmp("-film-size-t",argv[i]) ) {++i; time = atoi(argv[i]);}
		if( !strcmp("-film-exposure",argv[i]) ) {++i; time_resolution = atof(argv[i]);}
		if( !strcmp("-film-scanline",argv[i]) ) {++i; scanline = atoi(argv[i]);}
		if( !strcmp("-film-offset",argv[i]) ) {++i; film_offset = atof(argv[i]);}
		if( !strcmp("-film-single-pixel",argv[i]) ){++i; if( argc > i+1 ){ 	single_pixel = true; single_pixel_x = atoi(argv[i]), single_pixel_y = atoi(argv[i+1]); i+=1;}}
				
		if( !strcmp("-camera-position",argv[i]) ) {++i; if( argc > i+2 ){ camera_position = Vector3(atof(argv[i]),atof(argv[i+1]),atof(argv[i+2])); i+=2; }}
		if( !strcmp("-camera-focus",argv[i]) ) {++i; if( argc > i+2 ){ looking_at = Vector3(atof(argv[i]),atof(argv[i+1]),atof(argv[i+2])); i+=2;}}
		if( !strcmp("-camera-up",argv[i]) ) {++i; if( argc > i+2 ){ up = Vector3(atof(argv[i]),atof(argv[i+1]),atof(argv[i+2]));up.normalize(); i+=2; }}
		if( !strcmp("-camera-view-plane", argv[i]) ) {++i; view_plane_dist = atof(argv[i]);}
		if( !strcmp("-camera-spp",argv[i]) ) {++i; sqrt_spp = atoi(argv[i]);}
		if( !strcmp("-camera-focal-distance",argv[i])){++i; focal_distance = atof(argv[i]);}
		if( !strcmp("-camera-fov",argv[i])){++i; Real fov = M_PI/180.*atof(argv[i]); view_plane_dist = 1./tan(fov);}
		

		if( !strcmp("-cosine-light-source", argv[i]))
		{
			++i; 
			if( argc > i+6 )
			{	
				w->add_light(new CosineLightSource<3>( w, Vector3(atof(argv[i]),atof(argv[i+1]),atof(argv[i+2])), 
														  Vector3(atof(argv[i+3]),atof(argv[i+4]),atof(argv[i+5])),
														  Spectrum(atof(argv[i+6]))));
				
				i+=6;
			}
		}
		
		if( !strcmp("-point-light-source", argv[i]))
		{
			++i; 
			if( argc > i+3 )
			{	
				w->add_light(new PointLightSource<3>( w, Vector3(atof(argv[i]),atof(argv[i+1]),atof(argv[i+2])), 
														  Spectrum(atof(argv[i+3]))));
				
				i+=3;
			}
		}
				
		if( !strcmp("-spot-light-source", argv[i]))
		{
			++i; 
			if( argc > i+7 )
			{	
				w->add_light(new SpotLightSource<3>( w, Vector3(atof(argv[i]),atof(argv[i+1]),atof(argv[i+2])), 
														Vector3(atof(argv[i+3]),atof(argv[i+4]),atof(argv[i+5])),
														atof(argv[i+6]), Spectrum(atof(argv[i+7]))));
				i+=7;
			}
		}
		
		if( !strcmp("-rectangular-area-light-source", argv[i]))
		{
			++i; 
			if( argc > i+8 )
			{	
				Vector3 pos(atof(argv[i]),atof(argv[i+1]),atof(argv[i+2]));
				Vector3 dir(atof(argv[i+3]),atof(argv[i+4]),atof(argv[i+5])); 
				Vector2 size(atof(argv[i+6]),atof(argv[i+7]));
				Spectrum power(atof(argv[i+8]));

				i+=8;
					
				Imaging::Image *im = 0;
				if( argc > i+2 && !strcmp("-emission-texture",argv[i+1]) )
					im = new Imaging::Image(argv[i+2]);

				w->add_light(new RectangularAreaLightSource( w, pos, (dir-pos).normalize(), size, power, im ));
			}
		}


		if( !strcmp("-vpl-image-position", argv[i]) ){++i; image_coordinates_vpl = Vector2(atof(argv[i]),atof(argv[i+1])); i++;}
		
		if( !strcmp("-steady-state", argv[i])){transient = false;}
		if( !strcmp("-transient-state", argv[i])){transient = false;}

		if( !strcmp("-time-sampling",argv[i])) time_sampling = true;
		if( !strcmp("-standard-sampling", argv[i])){ time_sampling = false; }

		if( !strcmp("-log-name", argv[i]) ) {++i; log_file = argv[i];}
		if( !strcmp("-new-seed", argv[i]) ){++i; new_seed = atoi(argv[i]);}
		
		//PARTICIPATING MEDIA ARGUMENTS
		if (!strcmp("-homogeneous-medium", argv[i])){++i; medium_sigma_a = Spectrum(atof(argv[i])); ++i; medium_sigma_s = Spectrum(atof(argv[i]));}
		if (!strcmp("-hg", argv[i])) {i++; hg_g = atof(argv[i]);}
		if (!strcmp("-pf-time", argv[i])) {i++; pf_time = atof(argv[i]);}

		if (!strcmp("-max-nb-bounces", argv[i]))
		{	++i; max_nb_bounces = atoi(argv[i]);	}

		if (!strcmp("-scattering-level", argv[i])) 
		{
			++i;
			if (!strcmp("multiple", argv[i]))
				scat_level = Scattering::MULTIPLE;
			else if (!strcmp("single", argv[i]))
				scat_level = Scattering::SINGLE;
			else if (!strcmp("none", argv[i]))
				scat_level = Scattering::NONE;
			else 
				scat_level = Scattering::ALL;
		}
		
		if (!strcmp("-bidirectional-path-tracing", argv[i]))
		{
			printf("\nbidirectional_path_tracing\n\n");
			++i; vol_path_tracing_samples = atoi(argv[i]);
		}
		
		if (!strcmp("-multibounce-streak", argv[i])) {is_multibounce_streak = true; ++i; nb_multibounces = atoi(argv[i]); }
		if (!strcmp("-camera-unwarp", argv[i])) {camera_unwarp = true;}

		
	}
	
	//CREATE ALL FILENAMES INCLUDING FOLDER
	if (!log_file) 
	{
		log_file = (char*) malloc(sizeof(char) * 512);
		sprintf(log_file, "%s.log", name_file);
	}
	
	///////////////////////////////////////////////////////////////////////////////////////////////////////////
	FILE *f_log = fopen(log_file,"w+");
	//LOG FILM DATA
	fprintf(f_log, "FILM\n");
	fprintf(f_log, "====\n");
	fprintf(f_log, "Name: %s\n", name_file);
	fprintf(f_log, "Size(x-y-t): %d %d %d\n", width, height, time);
	fprintf(f_log, "Exposure: %f\n", time_resolution);
	fprintf(f_log, "Sqrt-spp: %d\n", sqrt_spp);
	fprintf(f_log, "Seed: %d\n", new_seed);
	if (sensor_mode)
		fprintf(f_log, "Camera-unwarp: --\n", time_resolution);
	else if (camera_unwarp)
		fprintf(f_log, "Camera-unwarp: TRUE\n", time_resolution);
	else
		fprintf(f_log, "Camera-unwarp: FALSE\n", time_resolution);

	fprintf(f_log, "\n");
	
	//LOG MEDIUM DATA
	fprintf(f_log, "MEDIUM\n");
	fprintf(f_log, "======\n");
	fprintf(f_log, "Absorption: %f %f %f\n", medium_sigma_a[0], medium_sigma_a[1], medium_sigma_a[2]);
	fprintf(f_log, "Scattering: %f %f %f\n", medium_sigma_s[0], medium_sigma_s[1], medium_sigma_s[2]);
	fprintf(f_log, "PF: g = %f\n", hg_g);
	switch (scat_level)
	{
		case Scattering::ALL: fprintf(f_log, "Scattering level: ALL\n"); break;
		case Scattering::SINGLE: fprintf(f_log, "Scattering level: SINGLE\n"); break;
		case Scattering::MULTIPLE: fprintf(f_log, "Scattering level: MULTI\n"); break;	
	}
	fprintf(f_log, "\n");

	///////////////////////////////////////////////////////////////////////////////////////////////////////////
	fprintf(f_log, "SCENE\n");
	fprintf(f_log, "=====\n");
		
	Camera3D *camera = 0;
	camera = new PinholeCamera(camera_position,looking_at,up,view_plane_dist);

	w->add_objects(objects);
	w->set_background(Spectrum(0.));
	
	// ----------------------------------------------------------------------
	// Create Film and Render engine
	Film *film = 0;
	StratifiedSampler sampler(width, height, sqrt_spp, true);
	
	RenderEngine<DIM> *engine; 
	
	if( transient )
	{
		// If rendering in transient state...
		// ...create streak-camera film...
		if (!is_multibounce_streak)
		{
			film = new StreakCameraFilm( width, height, time, time_resolution,  new BoxFilter(1.), camera_unwarp );

			static_cast<StreakCameraFilm*>(film)->set_name(name_file,"hdr");

			if (film_offset > 0.0) static_cast<StreakCameraFilm*>(film)->set_offset(film_offset);
		}
		else
		{
			film = new MultibounceStreakCameraFilm( width, height, time, time_resolution, nb_multibounces,
									new BoxFilter(1.), camera_unwarp );

			static_cast<MultibounceStreakCameraFilm*>(film)->set_name(name_file,"hdr");
		}

		
		// ... and the transient renderer.
		engine = new TransientRenderer<DIM>(f_log, time_sampling);
		static_cast<TransientRenderer<DIM>*>(engine)->set_sensor_mode(sensor_mode);
	}
	else
	{
		// If rendering in steady-state...
		// ... create regular film...
		film = new Film(width, height, (export_depth)?Film::ALL:Film::RADIANCE );
		// ... and steady-state renderer.
		engine = new RenderEngine<DIM>();
	}
	
	
	if( single_pixel && single_pixel_x > -1 && single_pixel_x < width
					 && single_pixel_y > -1 && single_pixel_y < height )
	{
		sampler.set_single_pixel(single_pixel_x, single_pixel_y);
		if( transient )
			static_cast<StreakCameraFilm*>(film)->set_streak(single_pixel_y);
	}

	timer.start();
	Real float_inf = std::numeric_limits<Real>::infinity();
	int int_inf = std::numeric_limits<int>::max();
	
	if( medium_sigma_a.avg() > 0.0001 && medium_sigma_s.avg() > 0.0001 )
		if (hg_g!=0.)
			w->set_medium(new HomogeneousMedium<DIM, PhaseFunction::HenyeyGreenstein<DIM> >(medium_sigma_a, medium_sigma_s, PhaseFunction::HenyeyGreenstein<DIM> (hg_g, pf_time)));
		else
			w->set_medium(new HomogeneousMedium<DIM, PhaseFunction::Isotropic<DIM> >(medium_sigma_a, medium_sigma_s, PhaseFunction::Isotropic<DIM> (pf_time)));
	
	srand(new_seed);

	// ----------------------------------------------------------------------
	// Create the integrator to be used
	BidirectionalPathTracing<DIM> bpt (w, vol_path_tracing_samples, f_log, film);

	bpt.set_path_tracing_only(true);
	if( transient )
	{
		bpt.set_max_path_size(max_nb_bounces);

		if( time_sampling )
			bpt.set_mode(1);
		else
			bpt.set_mode(0);
	}


	// ----------------------------------------------------------------------
	// GO RENDERING!
	printf("Rendering with Bidirectional Path Tracing...\n");
	engine->render( name_file, w, &bpt, camera, film, &sampler);


	//fflush(f_log);
	fclose(f_log);
	if( film )
		film->clear();
}